import logging
from typing import List
from typing import Optional
from typing import Union

from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import documented_by
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import format_freshness_table
from tecton._internals.utils import get_all_freshness
from tecton._internals.utils import is_live_workspace
from tecton.framework import data_source as framework_data_source
from tecton.framework import entity as framework_entity
from tecton.framework import feature_service as framework_feature_service
from tecton.framework import feature_view as framework_feature_view
from tecton.framework import transformation as framework_transformation
from tecton.framework.dataset import Dataset
from tecton.framework.dataset import LegacyDataset
from tecton.framework.dataset import SavedDataset
from tecton_core import conf
from tecton_core import specs
from tecton_core.fco_container import FcoContainer
from tecton_core.feature_definition_wrapper import FrameworkVersion
from tecton_proto.data import fco__client_pb2 as fco_pb2
from tecton_proto.metadataservice.metadata_service__client_pb2 import ArchiveSavedFeatureDataFrameRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetAllEntitiesRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetAllFeatureServicesRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetAllSavedFeatureDataFramesRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetAllTransformationsRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetAllVirtualDataSourcesRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetEntityRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetFeatureServiceRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetFeatureViewRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetSavedFeatureDataFrameRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetTransformationRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetVirtualDataSourceRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import ListWorkspacesRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import QueryFeatureViewsRequest


logger = logging.getLogger(__name__)

SUPPORTED_FRAMEWORK_VERSIONS = [FrameworkVersion.FWV5.value, FrameworkVersion.FWV6.value]


class Workspace:
    """
    This class represents a Workspace. The Workspace class is used to fetch Tecton Objects, which are stored in a Workspace.
    """

    def __init__(self, workspace: str, _is_live: Optional[bool] = None, _validate: bool = True):
        """
        Fetch an existing `tecton.Workspace` by name.

        :param workspace: Workspace name.
        :param _is_live: True if this workspace is live. Live workspaces materialize data for real-time endpoints,
         accessible in the production environment.
        :param _validate: If True, Tecton runs validation on object creation.
        """
        self.workspace = workspace

        if _is_live is None:
            self.is_live = is_live_workspace(self.workspace)
        else:
            self.is_live = _is_live

        if _validate:
            self._validate()

    def _validate(self):
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)

        workspace_from_resp = None
        for ws in response.workspaces:
            if ws.name == self.workspace:
                workspace_from_resp = ws
                break

        if workspace_from_resp is None:
            raise errors.NONEXISTENT_WORKSPACE(self.workspace, response.workspaces)

        if ws.capabilities.materializable != self.is_live:
            raise errors.INCORRECT_MATERIALIZATION_ENABLED_FLAG(self.is_live, ws.capabilities.materializable)

    def __repr__(self) -> str:
        capability_str = "Live" if self.is_live else "Development"
        return f"{self.workspace} ({capability_str})"

    @sdk_public_method
    def summary(self) -> Displayable:
        """Returns workspace metadata.

        :return: Displayable of metadata for this workspace. Use `to_dict()` for a parseable representation.
        """
        items = [
            ("Workspace Name", self.workspace),
            ("Workspace Type", "Live" if self.is_live else "Development"),
        ]
        return Displayable.from_properties(items=items)

    @classmethod
    @sdk_public_method
    def get(cls, name) -> "Workspace":
        """Returns a `tecton.Workspace` instance for the workspace with the provided name.

        :param name: The name of the workspace to retrieve.
        """
        return Workspace(name)

    @sdk_public_method
    def get_feature_view(self, name: str) -> framework_feature_view.FeatureView:
        """Returns a Feature View that has been applied to a workspace.

        :param name: The name of the Feature View to retrieve.
        """
        request = GetFeatureViewRequest(
            version_specifier=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetFeatureView(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        feature_view_spec = fco_container.get_single_root()

        if feature_view_spec is None:
            msg = f"Feature View '{name}' not found. Try running `workspace.list_feature_views()` to view all registered Feature Views."
            raise errors.TectonValidationError(msg)

        assert isinstance(feature_view_spec, specs.FeatureViewSpec)

        if isinstance(feature_view_spec, specs.FeatureTableSpec):
            msg = f"Feature View '{name}' not found. Did you mean workspace.get_feature_table('{name}')?"
            raise errors.TectonValidationError(msg)
        return framework_feature_view.feature_view_from_spec(feature_view_spec, fco_container)

    @sdk_public_method
    def get_feature_table(self, name: str) -> framework_feature_view.FeatureTable:
        """Returns a Feature Table that has been applied to a workspace.

        :param name: The name of the Feature Table to retrieve.
        """
        request = GetFeatureViewRequest(
            version_specifier=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetFeatureView(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        feature_table_spec = fco_container.get_single_root()

        if feature_table_spec is None:
            msg = f"Feature Table '{name}' not found. Try running `workspace.list_feature_tables()` to view all registered Feature Tables."
            raise errors.TectonValidationError(msg)

        assert isinstance(feature_table_spec, specs.FeatureViewSpec)
        if not isinstance(feature_table_spec, specs.FeatureTableSpec):
            msg = f"Feature Table '{name}' not found. Did you mean workspace.get_feature_view('{name}')?"
            raise errors.TectonValidationError(msg)

        if feature_table_spec.metadata.framework_version not in SUPPORTED_FRAMEWORK_VERSIONS:
            raise errors.UNSUPPORTED_FRAMEWORK_VERSION

        return framework_feature_view.FeatureTable._from_spec(feature_table_spec, fco_container)

    @sdk_public_method
    def get_prompt(self, name: str) -> framework_feature_view.Prompt:
        """Returns a Prompt that has been applied to a workspace.

        :param name: The name of the Prompt to retrieve.
        """
        request = GetFeatureViewRequest(
            version_specifier=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetFeatureView(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        prompt_spec = fco_container.get_single_root()

        if prompt_spec is None:
            msg = f"Prompt '{name}' not found. Try running `workspace.list_prompts()` to view all registered Prompts."
            raise errors.TectonValidationError(msg)

        if not isinstance(prompt_spec, specs.PromptSpec):
            msg = f"Prompt '{name}' not found. Did you mean workspace.get_feature_view('{name}')?"
            raise errors.TectonValidationError(msg)

        return framework_feature_view.Prompt._from_spec(prompt_spec, fco_container)

    @sdk_public_method
    def get_feature_service(self, name: str) -> framework_feature_service.FeatureService:
        """Returns a Feature Service that has been applied to a workspace.

        :param name: The name of the Feature Service to retrieve.
        """
        request = GetFeatureServiceRequest(
            service_reference=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetFeatureService(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        feature_service_spec = fco_container.get_single_root()

        if feature_service_spec is None:
            msg = f"Feature Service '{name}' not found. Try running `workspace.list_feature_services()` to view all registered Feature Services."
            raise errors.TectonValidationError(msg)

        assert isinstance(feature_service_spec, specs.FeatureServiceSpec)

        return framework_feature_service.FeatureService._from_spec(feature_service_spec, fco_container)

    @sdk_public_method
    def get_data_source(
        self, name: str
    ) -> Union[
        framework_data_source.BatchSource,
        framework_data_source.StreamSource,
    ]:
        """Returns a Data Source that has been applied to a workspace.

        :param name: The name of the Data Source to retrieve.
        """
        request = GetVirtualDataSourceRequest(
            name=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetVirtualDataSource(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        data_source_spec = fco_container.get_single_root()

        if data_source_spec is None:
            msg = f"Data Source '{name}' not found. Try running `workspace.list_data_sources()` to view all registered Data Sources."
            raise errors.TectonValidationError(msg)

        assert isinstance(data_source_spec, specs.DataSourceSpec)

        if data_source_spec.metadata.framework_version not in SUPPORTED_FRAMEWORK_VERSIONS:
            raise errors.UNSUPPORTED_FRAMEWORK_VERSION

        return framework_data_source.data_source_from_spec(data_source_spec)

    @sdk_public_method
    def get_entity(self, name: str) -> framework_entity.Entity:
        """Returns an Entity that has been applied to a workspace.

        :param name: The name of the Entity to retrieve.
        """
        request = GetEntityRequest(
            name=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetEntity(request)

        fco_container = FcoContainer.from_proto(response.fco_container)
        entity_spec = fco_container.get_single_root()

        if entity_spec is None:
            msg = f"Entity '{name}' not found. Try running `workspace.list_entities()` to view all registered Entities."
            raise errors.TectonValidationError(msg)

        assert isinstance(entity_spec, specs.EntitySpec)

        return framework_entity.Entity._from_spec(entity_spec)

    @sdk_public_method
    def get_transformation(self, name: str) -> framework_transformation.Transformation:
        """Returns a Transformation that has been applied to a workspace.

        :param name: The name of the Transformation to retrieve.
        """
        request = GetTransformationRequest(
            name=name,
            workspace=self.workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetTransformation(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        transformation_spec = fco_container.get_single_root()

        if transformation_spec is None:
            msg = f"Transformation '{name}' not found. Try running `workspace.list_transformations()` to view all registered Transformations."
            raise errors.TectonValidationError(msg)

        assert isinstance(transformation_spec, specs.TransformationSpec)
        return framework_transformation.Transformation._from_spec(transformation_spec)

    @sdk_public_method
    def get_dataset(self, name: str) -> Dataset:
        """Returns a Dataset that has been saved to this workspace.

        :param name: The name of the Dataset to retrieve.
        """
        request = GetSavedFeatureDataFrameRequest(saved_feature_dataframe_name=name, workspace=self.workspace)
        response = metadata_service.instance().GetSavedFeatureDataFrame(request)
        if response.saved_feature_dataframe.WhichOneof("kind") == "saved_dataset":
            return SavedDataset._from_proto(response.saved_feature_dataframe)
        else:
            return LegacyDataset._from_proto(response.saved_feature_dataframe)

    @sdk_public_method
    def delete_dataset(self, name: str):
        """Deletes a Dataset that has been saved to this workspace.

        :param name: The name of the Dataset to delete.
        """
        request = GetSavedFeatureDataFrameRequest(saved_feature_dataframe_name=name, workspace=self.workspace)
        response = metadata_service.instance().GetSavedFeatureDataFrame(request)

        request = ArchiveSavedFeatureDataFrameRequest()
        request.saved_feature_dataframe_id.CopyFrom(response.saved_feature_dataframe.saved_feature_dataframe_id)
        metadata_service.instance().ArchiveSavedFeatureDataFrame(request)
        print(f"Dataset {response.saved_feature_dataframe.info.name} deleted")

    @sdk_public_method
    def list_datasets(self) -> List[str]:
        """Returns a list of all saved Datasets within a workspace.

        :return: List of strings of dataset names
        """
        request = GetAllSavedFeatureDataFramesRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllSavedFeatureDataFrames(request)
        return sorted([sfdf.info.name for sfdf in response.saved_feature_dataframes])

    @sdk_public_method
    def list_feature_views(self) -> List[str]:
        """Returns a list of all registered Feature Views within a workspace.

        :return: List of strings of feature view names
        """

        def feature_view_filter(fco: fco_pb2.Fco) -> bool:
            return fco.HasField("feature_view") and not (
                fco.feature_view.HasField("feature_table") or fco.feature_view.HasField("prompt")
            )

        return self._list_filtered_feature_views(feature_view_filter)

    @sdk_public_method
    def list_prompts(self) -> List[str]:
        """Returns a list of all registered Prompts within a workspace.

        :return: List of strings of prompt names
        """

        def prompt_filter(fco: fco_pb2.Fco) -> bool:
            return fco.HasField("feature_view") and fco.feature_view.HasField("prompt")

        return self._list_filtered_feature_views(prompt_filter)

    @sdk_public_method
    def list_feature_services(self) -> List[str]:
        """Returns a list of all registered Feature Services within a workspace.

        :return: List of strings of feature service names
        """
        request = GetAllFeatureServicesRequest()
        request.workspace = self.workspace
        response = metadata_service.instance().GetAllFeatureServices(request)
        return sorted([proto.fco_metadata.name for proto in response.feature_services])

    @sdk_public_method
    def list_transformations(self) -> List[str]:
        """Returns a list of all registered Transformations within a workspace.

        :return: List of strings of transformation names
        """
        request = GetAllTransformationsRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllTransformations(request)
        return sorted([proto.fco_metadata.name for proto in response.transformations])

    @sdk_public_method
    def list_entities(self) -> List[str]:
        """Returns a list of all registered Entities within a workspace.

        :return: List of strings of entity names
        """
        request = GetAllEntitiesRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllEntities(request)
        return sorted([proto.fco_metadata.name for proto in response.entities])

    @sdk_public_method
    def list_data_sources(self) -> List[str]:
        """Returns a list of all registered Data Sources within a workspace.

        :return: List of strings of data source names
        """
        request = GetAllVirtualDataSourcesRequest(workspace=self.workspace)
        response = metadata_service.instance().GetAllVirtualDataSources(request)
        return sorted([proto.fco_metadata.name for proto in response.virtual_data_sources])

    @sdk_public_method
    def list_feature_tables(self) -> List[str]:
        """Returns a list of all registered Feature Tables within a workspace.

        :return: List of strings of feature table names
        """

        def feature_table_filter(fco: fco_pb2.Fco) -> bool:
            return fco.HasField("feature_view") and fco.feature_view.HasField("feature_table")

        return self._list_filtered_feature_views(feature_table_filter)

    def _list_filtered_feature_views(self, filter_function):
        request = QueryFeatureViewsRequest(workspace=self.workspace)
        response = metadata_service.instance().QueryFeatureViews(request)

        # Do not extract these data protos into "specs" since that runs function deserialization and other work which
        # is not needed here and may throw errors if there is a bad definition.
        fv_fcos = filter(filter_function, response.fco_container.fcos)
        return sorted([fco.feature_view.fco_metadata.name for fco in fv_fcos])

    @sdk_public_method
    def get_feature_freshness(self) -> Union[Displayable, str]:
        """Returns feature freshness status for Feature Views and Tables.

        :return: Displayable containing freshness statuses for all features. Use `to_dict()` for a parseable representation.
        """
        freshness_statuses = get_all_freshness(self.workspace)

        if len(freshness_statuses) == 0:
            return "No Feature Views found in this workspace"
        return format_freshness_table(freshness_statuses)


@sdk_public_method
@documented_by(Workspace.get)
def get_workspace(name: str) -> Workspace:
    return Workspace.get(name)


@sdk_public_method
def get_feature_service(name: str, workspace: str) -> framework_feature_service.FeatureService:
    """Returns a Feature Service that has been applied to a workspace.

    :param name: The name of the Feature Service to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_feature_service(name)


@sdk_public_method
def get_feature_table(name: str, workspace: str) -> framework_feature_view.FeatureTable:
    """Returns a Feature Table that has been applied to a workspace.

    :param name: The name of the Feature Table to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_feature_table(name)


@sdk_public_method
def get_feature_view(name: str, workspace: str) -> framework_feature_view.FeatureView:
    """Returns a Feature View that has been applied to a workspace.

    :param name: The name of the Feature View to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_feature_view(name)


@sdk_public_method
def get_prompt(name: str, workspace: str) -> framework_feature_view.FeatureTable:
    """Returns a Prompt that has been applied to a workspace.

    :param name: The name of the Prompt to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_prompt(name)


@sdk_public_method
def get_entity(name: str, workspace: str) -> framework_entity.Entity:
    """Returns an Entity that has been applied to a workspace.

    :param name: The name of the Entity to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_entity(name)


@sdk_public_method
def get_transformation(name: str, workspace: str) -> framework_transformation.Transformation:
    """Returns a Transformation that has been applied to a workspace.

    :param name: The name of the Transformation to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_transformation(name)


@sdk_public_method
def get_data_source(
    name: str, workspace: str
) -> Union[
    framework_data_source.BatchSource,
    framework_data_source.StreamSource,
]:
    """Returns a Data Source that has been applied to a workspace.

    :param name: The name of the Data Source to retrieve.
    :param workspace: The name of the workspace.
    """
    return get_workspace(workspace).get_data_source(name)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.auth import principal__client_pb2 as _principal__client_pb2
from tecton_proto.auth import service__client_pb2 as _service__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Text, Union

DELETED: ModelArtifactStatus
DESCRIPTOR: _descriptor.FileDescriptor
ERROR: ModelArtifactStatus
MODEL_TYPE_UNSPECIFIED: ModelType
PENDING_FILE: ModelArtifactStatus
PENDING_SCAN: ModelArtifactStatus
PYTORCH: ModelType
READY: ModelArtifactStatus
STATUS_UNSPECIFIED: ModelArtifactStatus
TECTON_TEXT_EMBEDDING: ModelType

class CompleteModelArtifactUploadRequest(_message.Message):
    __slots__ = ["model_artifact_id", "part_etags", "upload_id"]
    class PartEtagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: int
        value: str
        def __init__(self, key: Optional[int] = ..., value: Optional[str] = ...) -> None: ...
    MODEL_ARTIFACT_ID_FIELD_NUMBER: ClassVar[int]
    PART_ETAGS_FIELD_NUMBER: ClassVar[int]
    UPLOAD_ID_FIELD_NUMBER: ClassVar[int]
    model_artifact_id: _id__client_pb2.Id
    part_etags: _containers.ScalarMap[int, str]
    upload_id: str
    def __init__(self, model_artifact_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., upload_id: Optional[str] = ..., part_etags: Optional[Mapping[int, str]] = ...) -> None: ...

class CompleteModelArtifactUploadResponse(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: ClassVar[int]
    key: str
    def __init__(self, key: Optional[str] = ...) -> None: ...

class CreateModelArtifactRequest(_message.Message):
    __slots__ = ["artifact_files", "description", "environments", "file_hashes", "input_schema", "model_config_file_path", "model_file_path", "name", "output_schema", "tags", "type"]
    class FileHashesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ARTIFACT_FILES_FIELD_NUMBER: ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: ClassVar[int]
    FILE_HASHES_FIELD_NUMBER: ClassVar[int]
    INPUT_SCHEMA_FIELD_NUMBER: ClassVar[int]
    MODEL_CONFIG_FILE_PATH_FIELD_NUMBER: ClassVar[int]
    MODEL_FILE_PATH_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    OUTPUT_SCHEMA_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    artifact_files: _containers.RepeatedScalarFieldContainer[str]
    description: str
    environments: _containers.RepeatedScalarFieldContainer[str]
    file_hashes: _containers.ScalarMap[str, str]
    input_schema: _schema__client_pb2.Schema
    model_config_file_path: str
    model_file_path: str
    name: str
    output_schema: _schema__client_pb2.Schema
    tags: _containers.ScalarMap[str, str]
    type: ModelType
    def __init__(self, name: Optional[str] = ..., type: Optional[Union[ModelType, str]] = ..., description: Optional[str] = ..., file_hashes: Optional[Mapping[str, str]] = ..., input_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., output_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., tags: Optional[Mapping[str, str]] = ..., model_file_path: Optional[str] = ..., model_config_file_path: Optional[str] = ..., artifact_files: Optional[Iterable[str]] = ..., environments: Optional[Iterable[str]] = ...) -> None: ...

class CreateModelArtifactResponse(_message.Message):
    __slots__ = ["model_artifact_info"]
    MODEL_ARTIFACT_INFO_FIELD_NUMBER: ClassVar[int]
    model_artifact_info: ModelArtifactInfo
    def __init__(self, model_artifact_info: Optional[Union[ModelArtifactInfo, Mapping]] = ...) -> None: ...

class DeleteModelArtifactRequest(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    id: _id__client_pb2.Id
    name: str
    def __init__(self, id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ...) -> None: ...

class DeleteModelArtifactResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class DescribeModelArtifactRequest(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    id: _id__client_pb2.Id
    name: str
    def __init__(self, id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ...) -> None: ...

class DescribeModelArtifactResponse(_message.Message):
    __slots__ = ["model_artifact_info"]
    MODEL_ARTIFACT_INFO_FIELD_NUMBER: ClassVar[int]
    model_artifact_info: ModelArtifactInfo
    def __init__(self, model_artifact_info: Optional[Union[ModelArtifactInfo, Mapping]] = ...) -> None: ...

class FetchModelArtifactRequest(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    id: _id__client_pb2.Id
    name: str
    def __init__(self, id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ...) -> None: ...

class FetchModelArtifactResponse(_message.Message):
    __slots__ = ["model_artifact_download_url", "model_config_download_url"]
    MODEL_ARTIFACT_DOWNLOAD_URL_FIELD_NUMBER: ClassVar[int]
    MODEL_CONFIG_DOWNLOAD_URL_FIELD_NUMBER: ClassVar[int]
    model_artifact_download_url: str
    model_config_download_url: str
    def __init__(self, model_artifact_download_url: Optional[str] = ..., model_config_download_url: Optional[str] = ...) -> None: ...

class GetModelArtifactUploadUrlRequest(_message.Message):
    __slots__ = ["model_artifact_id"]
    MODEL_ARTIFACT_ID_FIELD_NUMBER: ClassVar[int]
    model_artifact_id: _id__client_pb2.Id
    def __init__(self, model_artifact_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ...) -> None: ...

class GetModelArtifactUploadUrlResponse(_message.Message):
    __slots__ = ["model_artifact_id", "model_config_upload_url", "upload_id"]
    MODEL_ARTIFACT_ID_FIELD_NUMBER: ClassVar[int]
    MODEL_CONFIG_UPLOAD_URL_FIELD_NUMBER: ClassVar[int]
    UPLOAD_ID_FIELD_NUMBER: ClassVar[int]
    model_artifact_id: _id__client_pb2.Id
    model_config_upload_url: str
    upload_id: str
    def __init__(self, model_artifact_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., upload_id: Optional[str] = ..., model_config_upload_url: Optional[str] = ...) -> None: ...

class GetTectonModelInfoRequest(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: ClassVar[int]
    name: str
    def __init__(self, name: Optional[str] = ...) -> None: ...

class GetTectonModelInfoResponse(_message.Message):
    __slots__ = ["model_info"]
    MODEL_INFO_FIELD_NUMBER: ClassVar[int]
    model_info: ModelInfo
    def __init__(self, model_info: Optional[Union[ModelInfo, Mapping]] = ...) -> None: ...

class ListModelArtifactsRequest(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    id: _id__client_pb2.Id
    name: str
    def __init__(self, id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ...) -> None: ...

class ListModelArtifactsResponse(_message.Message):
    __slots__ = ["models"]
    MODELS_FIELD_NUMBER: ClassVar[int]
    models: _containers.RepeatedCompositeFieldContainer[ModelArtifactInfo]
    def __init__(self, models: Optional[Iterable[Union[ModelArtifactInfo, Mapping]]] = ...) -> None: ...

class ListTectonModelsRequest(_message.Message):
    __slots__ = ["filter"]
    FILTER_FIELD_NUMBER: ClassVar[int]
    filter: str
    def __init__(self, filter: Optional[str] = ...) -> None: ...

class ListTectonModelsResponse(_message.Message):
    __slots__ = ["models"]
    MODELS_FIELD_NUMBER: ClassVar[int]
    models: _containers.RepeatedCompositeFieldContainer[ModelInfo]
    def __init__(self, models: Optional[Iterable[Union[ModelInfo, Mapping]]] = ...) -> None: ...

class ModelArtifactInfo(_message.Message):
    __slots__ = ["artifact_files", "created_at", "created_by", "created_by_principal", "description", "environments", "file_hashes", "id", "input_schema", "model_config_file_path", "model_file_path", "name", "output_schema", "status", "storage_path", "tags", "type", "updated_at"]
    class FileHashesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ARTIFACT_FILES_FIELD_NUMBER: ClassVar[int]
    CREATED_AT_FIELD_NUMBER: ClassVar[int]
    CREATED_BY_FIELD_NUMBER: ClassVar[int]
    CREATED_BY_PRINCIPAL_FIELD_NUMBER: ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: ClassVar[int]
    FILE_HASHES_FIELD_NUMBER: ClassVar[int]
    ID_FIELD_NUMBER: ClassVar[int]
    INPUT_SCHEMA_FIELD_NUMBER: ClassVar[int]
    MODEL_CONFIG_FILE_PATH_FIELD_NUMBER: ClassVar[int]
    MODEL_FILE_PATH_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    OUTPUT_SCHEMA_FIELD_NUMBER: ClassVar[int]
    STATUS_FIELD_NUMBER: ClassVar[int]
    STORAGE_PATH_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: ClassVar[int]
    artifact_files: _containers.RepeatedScalarFieldContainer[str]
    created_at: _timestamp_pb2.Timestamp
    created_by: _principal__client_pb2.Principal
    created_by_principal: _principal__client_pb2.PrincipalBasic
    description: str
    environments: _containers.RepeatedScalarFieldContainer[str]
    file_hashes: _containers.ScalarMap[str, str]
    id: _id__client_pb2.Id
    input_schema: _schema__client_pb2.Schema
    model_config_file_path: str
    model_file_path: str
    name: str
    output_schema: _schema__client_pb2.Schema
    status: ModelArtifactStatus
    storage_path: str
    tags: _containers.ScalarMap[str, str]
    type: ModelType
    updated_at: _timestamp_pb2.Timestamp
    def __init__(self, id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ..., type: Optional[Union[ModelType, str]] = ..., description: Optional[str] = ..., file_hashes: Optional[Mapping[str, str]] = ..., input_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., output_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., tags: Optional[Mapping[str, str]] = ..., storage_path: Optional[str] = ..., status: Optional[Union[ModelArtifactStatus, str]] = ..., created_at: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., updated_at: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., model_file_path: Optional[str] = ..., model_config_file_path: Optional[str] = ..., artifact_files: Optional[Iterable[str]] = ..., environments: Optional[Iterable[str]] = ..., created_by: Optional[Union[_principal__client_pb2.Principal, Mapping]] = ..., created_by_principal: Optional[Union[_principal__client_pb2.PrincipalBasic, Mapping]] = ...) -> None: ...

class ModelInfo(_message.Message):
    __slots__ = ["id", "metadata_public_uri", "model_public_uri", "name"]
    ID_FIELD_NUMBER: ClassVar[int]
    METADATA_PUBLIC_URI_FIELD_NUMBER: ClassVar[int]
    MODEL_PUBLIC_URI_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    id: _id__client_pb2.Id
    metadata_public_uri: str
    model_public_uri: str
    name: str
    def __init__(self, id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ..., model_public_uri: Optional[str] = ..., metadata_public_uri: Optional[str] = ...) -> None: ...

class UpdateTectonModelRequest(_message.Message):
    __slots__ = ["artifact_files", "description", "environments", "id", "input_schema", "model_config_file_path", "model_file_path", "name", "output_schema", "tags", "type"]
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ARTIFACT_FILES_FIELD_NUMBER: ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: ClassVar[int]
    ID_FIELD_NUMBER: ClassVar[int]
    INPUT_SCHEMA_FIELD_NUMBER: ClassVar[int]
    MODEL_CONFIG_FILE_PATH_FIELD_NUMBER: ClassVar[int]
    MODEL_FILE_PATH_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    OUTPUT_SCHEMA_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    artifact_files: _containers.RepeatedScalarFieldContainer[str]
    description: str
    environments: _containers.RepeatedScalarFieldContainer[str]
    id: _id__client_pb2.Id
    input_schema: _schema__client_pb2.Schema
    model_config_file_path: str
    model_file_path: str
    name: str
    output_schema: _schema__client_pb2.Schema
    tags: _containers.ScalarMap[str, str]
    type: ModelType
    def __init__(self, name: Optional[str] = ..., id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., type: Optional[Union[ModelType, str]] = ..., description: Optional[str] = ..., input_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., output_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., tags: Optional[Mapping[str, str]] = ..., model_file_path: Optional[str] = ..., model_config_file_path: Optional[str] = ..., artifact_files: Optional[Iterable[str]] = ..., environments: Optional[Iterable[str]] = ...) -> None: ...

class UpdateTectonModelResponse(_message.Message):
    __slots__ = ["model_config_upload_url"]
    MODEL_CONFIG_UPLOAD_URL_FIELD_NUMBER: ClassVar[int]
    model_config_upload_url: str
    def __init__(self, model_config_upload_url: Optional[str] = ...) -> None: ...

class UploadModelArtifactPartRequest(_message.Message):
    __slots__ = ["model_artifact_id", "parent_upload_id", "part_number"]
    MODEL_ARTIFACT_ID_FIELD_NUMBER: ClassVar[int]
    PARENT_UPLOAD_ID_FIELD_NUMBER: ClassVar[int]
    PART_NUMBER_FIELD_NUMBER: ClassVar[int]
    model_artifact_id: _id__client_pb2.Id
    parent_upload_id: str
    part_number: int
    def __init__(self, model_artifact_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., parent_upload_id: Optional[str] = ..., part_number: Optional[int] = ...) -> None: ...

class UploadModelArtifactPartResponse(_message.Message):
    __slots__ = ["upload_url"]
    UPLOAD_URL_FIELD_NUMBER: ClassVar[int]
    upload_url: str
    def __init__(self, upload_url: Optional[str] = ...) -> None: ...

class ModelArtifactStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ModelType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton_proto.common import container_image__client_pb2 as _container_image__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class AWSInstanceGroup(_message.Message):
    __slots__ = ["ami_image_id", "autoscaling_group_arn", "autoscaling_group_name", "health_check_path", "iam_instance_profile_arn", "instance_type", "launch_template_id", "port", "region", "security_group_ids", "subnet_ids"]
    AMI_IMAGE_ID_FIELD_NUMBER: ClassVar[int]
    AUTOSCALING_GROUP_ARN_FIELD_NUMBER: ClassVar[int]
    AUTOSCALING_GROUP_NAME_FIELD_NUMBER: ClassVar[int]
    HEALTH_CHECK_PATH_FIELD_NUMBER: ClassVar[int]
    IAM_INSTANCE_PROFILE_ARN_FIELD_NUMBER: ClassVar[int]
    INSTANCE_TYPE_FIELD_NUMBER: ClassVar[int]
    LAUNCH_TEMPLATE_ID_FIELD_NUMBER: ClassVar[int]
    PORT_FIELD_NUMBER: ClassVar[int]
    REGION_FIELD_NUMBER: ClassVar[int]
    SECURITY_GROUP_IDS_FIELD_NUMBER: ClassVar[int]
    SUBNET_IDS_FIELD_NUMBER: ClassVar[int]
    ami_image_id: str
    autoscaling_group_arn: str
    autoscaling_group_name: str
    health_check_path: str
    iam_instance_profile_arn: str
    instance_type: str
    launch_template_id: str
    port: int
    region: str
    security_group_ids: _containers.RepeatedScalarFieldContainer[str]
    subnet_ids: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, autoscaling_group_arn: Optional[str] = ..., autoscaling_group_name: Optional[str] = ..., region: Optional[str] = ..., port: Optional[int] = ..., health_check_path: Optional[str] = ..., instance_type: Optional[str] = ..., ami_image_id: Optional[str] = ..., iam_instance_profile_arn: Optional[str] = ..., security_group_ids: Optional[Iterable[str]] = ..., subnet_ids: Optional[Iterable[str]] = ..., launch_template_id: Optional[str] = ...) -> None: ...

class AWSInstanceGroupUpdateConfig(_message.Message):
    __slots__ = ["ami_image_id", "instance_type"]
    AMI_IMAGE_ID_FIELD_NUMBER: ClassVar[int]
    INSTANCE_TYPE_FIELD_NUMBER: ClassVar[int]
    ami_image_id: str
    instance_type: str
    def __init__(self, instance_type: Optional[str] = ..., ami_image_id: Optional[str] = ...) -> None: ...

class AWSTargetGroup(_message.Message):
    __slots__ = ["arn", "instance_group", "name"]
    ARN_FIELD_NUMBER: ClassVar[int]
    INSTANCE_GROUP_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    arn: str
    instance_group: AWSInstanceGroup
    name: str
    def __init__(self, arn: Optional[str] = ..., name: Optional[str] = ..., instance_group: Optional[Union[AWSInstanceGroup, Mapping]] = ...) -> None: ...

class CapacityConfig(_message.Message):
    __slots__ = ["autoscaling_enabled", "desired_nodes", "max_nodes", "min_nodes"]
    AUTOSCALING_ENABLED_FIELD_NUMBER: ClassVar[int]
    DESIRED_NODES_FIELD_NUMBER: ClassVar[int]
    MAX_NODES_FIELD_NUMBER: ClassVar[int]
    MIN_NODES_FIELD_NUMBER: ClassVar[int]
    autoscaling_enabled: bool
    desired_nodes: int
    max_nodes: int
    min_nodes: int
    def __init__(self, autoscaling_enabled: bool = ..., min_nodes: Optional[int] = ..., max_nodes: Optional[int] = ..., desired_nodes: Optional[int] = ...) -> None: ...

class GoogleCloudBackendService(_message.Message):
    __slots__ = ["instance_group", "project", "region", "target_id"]
    INSTANCE_GROUP_FIELD_NUMBER: ClassVar[int]
    PROJECT_FIELD_NUMBER: ClassVar[int]
    REGION_FIELD_NUMBER: ClassVar[int]
    TARGET_ID_FIELD_NUMBER: ClassVar[int]
    instance_group: GoogleCloudInstanceGroup
    project: str
    region: str
    target_id: str
    def __init__(self, target_id: Optional[str] = ..., project: Optional[str] = ..., region: Optional[str] = ..., instance_group: Optional[Union[GoogleCloudInstanceGroup, Mapping]] = ...) -> None: ...

class GoogleCloudInstanceGroup(_message.Message):
    __slots__ = ["health_check_name", "machine_type", "project", "region", "subnetworks", "target_id"]
    HEALTH_CHECK_NAME_FIELD_NUMBER: ClassVar[int]
    MACHINE_TYPE_FIELD_NUMBER: ClassVar[int]
    PROJECT_FIELD_NUMBER: ClassVar[int]
    REGION_FIELD_NUMBER: ClassVar[int]
    SUBNETWORKS_FIELD_NUMBER: ClassVar[int]
    TARGET_ID_FIELD_NUMBER: ClassVar[int]
    health_check_name: str
    machine_type: str
    project: str
    region: str
    subnetworks: _containers.RepeatedScalarFieldContainer[str]
    target_id: str
    def __init__(self, project: Optional[str] = ..., region: Optional[str] = ..., target_id: Optional[str] = ..., machine_type: Optional[str] = ..., subnetworks: Optional[Iterable[str]] = ..., health_check_name: Optional[str] = ...) -> None: ...

class HealthCheckConfig(_message.Message):
    __slots__ = ["path", "port", "protocol_type"]
    PATH_FIELD_NUMBER: ClassVar[int]
    PORT_FIELD_NUMBER: ClassVar[int]
    PROTOCOL_TYPE_FIELD_NUMBER: ClassVar[int]
    path: str
    port: NamedPort
    protocol_type: str
    def __init__(self, port: Optional[Union[NamedPort, Mapping]] = ..., path: Optional[str] = ..., protocol_type: Optional[str] = ...) -> None: ...

class InstanceGroup(_message.Message):
    __slots__ = ["aws_instance_group", "capacity", "container_image", "environment_variables", "google_cloud_instance_group", "grpc_port", "health_check_config", "health_check_name", "http_port", "name", "prometheus_port", "tags"]
    class EnvironmentVariablesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    AWS_INSTANCE_GROUP_FIELD_NUMBER: ClassVar[int]
    CAPACITY_FIELD_NUMBER: ClassVar[int]
    CONTAINER_IMAGE_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENT_VARIABLES_FIELD_NUMBER: ClassVar[int]
    GOOGLE_CLOUD_INSTANCE_GROUP_FIELD_NUMBER: ClassVar[int]
    GRPC_PORT_FIELD_NUMBER: ClassVar[int]
    HEALTH_CHECK_CONFIG_FIELD_NUMBER: ClassVar[int]
    HEALTH_CHECK_NAME_FIELD_NUMBER: ClassVar[int]
    HTTP_PORT_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    PROMETHEUS_PORT_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    aws_instance_group: AWSInstanceGroup
    capacity: CapacityConfig
    container_image: _container_image__client_pb2.ContainerImage
    environment_variables: _containers.ScalarMap[str, str]
    google_cloud_instance_group: GoogleCloudInstanceGroup
    grpc_port: NamedPort
    health_check_config: HealthCheckConfig
    health_check_name: str
    http_port: NamedPort
    name: str
    prometheus_port: NamedPort
    tags: _containers.ScalarMap[str, str]
    def __init__(self, name: Optional[str] = ..., container_image: Optional[Union[_container_image__client_pb2.ContainerImage, Mapping]] = ..., capacity: Optional[Union[CapacityConfig, Mapping]] = ..., health_check_config: Optional[Union[HealthCheckConfig, Mapping]] = ..., health_check_name: Optional[str] = ..., prometheus_port: Optional[Union[NamedPort, Mapping]] = ..., grpc_port: Optional[Union[NamedPort, Mapping]] = ..., http_port: Optional[Union[NamedPort, Mapping]] = ..., aws_instance_group: Optional[Union[AWSInstanceGroup, Mapping]] = ..., google_cloud_instance_group: Optional[Union[GoogleCloudInstanceGroup, Mapping]] = ..., tags: Optional[Mapping[str, str]] = ..., environment_variables: Optional[Mapping[str, str]] = ...) -> None: ...

class InstanceGroupHandle(_message.Message):
    __slots__ = ["instance_group_id", "instance_group_name", "instance_group_template_id"]
    INSTANCE_GROUP_ID_FIELD_NUMBER: ClassVar[int]
    INSTANCE_GROUP_NAME_FIELD_NUMBER: ClassVar[int]
    INSTANCE_GROUP_TEMPLATE_ID_FIELD_NUMBER: ClassVar[int]
    instance_group_id: str
    instance_group_name: str
    instance_group_template_id: str
    def __init__(self, instance_group_id: Optional[str] = ..., instance_group_name: Optional[str] = ..., instance_group_template_id: Optional[str] = ...) -> None: ...

class InstanceGroupStatus(_message.Message):
    __slots__ = ["healthy_instances"]
    HEALTHY_INSTANCES_FIELD_NUMBER: ClassVar[int]
    healthy_instances: int
    def __init__(self, healthy_instances: Optional[int] = ...) -> None: ...

class LoadBalancerTarget(_message.Message):
    __slots__ = ["aws_target_group", "google_backend_service"]
    AWS_TARGET_GROUP_FIELD_NUMBER: ClassVar[int]
    GOOGLE_BACKEND_SERVICE_FIELD_NUMBER: ClassVar[int]
    aws_target_group: AWSTargetGroup
    google_backend_service: GoogleCloudBackendService
    def __init__(self, aws_target_group: Optional[Union[AWSTargetGroup, Mapping]] = ..., google_backend_service: Optional[Union[GoogleCloudBackendService, Mapping]] = ...) -> None: ...

class NamedPort(_message.Message):
    __slots__ = ["port_name", "port_number"]
    PORT_NAME_FIELD_NUMBER: ClassVar[int]
    PORT_NUMBER_FIELD_NUMBER: ClassVar[int]
    port_name: str
    port_number: int
    def __init__(self, port_number: Optional[int] = ..., port_name: Optional[str] = ...) -> None: ...



from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Optional

DESCRIPTOR: _descriptor.FileDescriptor

class UserDefinedFunction(_message.Message):
    __slots__ = ["body", "name"]
    BODY_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    body: str
    name: str
    def __init__(self, name: Optional[str] = ..., body: Optional[str] = ...) -> None: ...



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar

COLUMN_TYPE_BOOL: ColumnType
COLUMN_TYPE_DERIVE_FROM_DATA_TYPE: ColumnType
COLUMN_TYPE_DOUBLE: ColumnType
COLUMN_TYPE_DOUBLE_ARRAY: ColumnType
COLUMN_TYPE_FLOAT_ARRAY: ColumnType
COLUMN_TYPE_INT64: ColumnType
COLUMN_TYPE_INT64_ARRAY: ColumnType
COLUMN_TYPE_STRING: ColumnType
COLUMN_TYPE_STRING_ARRAY: ColumnType
COLUMN_TYPE_UNKNOWN: ColumnType
DESCRIPTOR: _descriptor.FileDescriptor

class ColumnType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton_proto.spark_api import jobs__client_pb2 as _jobs__client_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Mapping, Optional, Union

BATCH: TaskType
DATASET_GENERATION: TaskType
DELETION: TaskType
DELTA_MAINTENANCE: TaskType
DESCRIPTOR: _descriptor.FileDescriptor
ENV_DATABRICKS_NOTEBOOK: SparkExecutionEnvironment
ENV_DATAPROC: SparkExecutionEnvironment
ENV_EMR: SparkExecutionEnvironment
ENV_UNSPECIFIED: SparkExecutionEnvironment
FEATURE_EXPORT: TaskType
INGEST: TaskType
PLAN_INTEGRATION_TEST_BATCH: TaskType
PLAN_INTEGRATION_TEST_STREAM: TaskType
STREAMING: TaskType
UNKNOWN: TaskType

class JobRequestTemplates(_message.Message):
    __slots__ = ["databricks_template", "emr_template"]
    DATABRICKS_TEMPLATE_FIELD_NUMBER: ClassVar[int]
    EMR_TEMPLATE_FIELD_NUMBER: ClassVar[int]
    databricks_template: _jobs__client_pb2.StartJobRequest
    emr_template: _jobs__client_pb2.StartJobRequest
    def __init__(self, databricks_template: Optional[Union[_jobs__client_pb2.StartJobRequest, Mapping]] = ..., emr_template: Optional[Union[_jobs__client_pb2.StartJobRequest, Mapping]] = ...) -> None: ...

class SparkClusterEnvironment(_message.Message):
    __slots__ = ["job_request_templates", "merged_user_deployment_settings_version", "spark_cluster_environment_version"]
    JOB_REQUEST_TEMPLATES_FIELD_NUMBER: ClassVar[int]
    MERGED_USER_DEPLOYMENT_SETTINGS_VERSION_FIELD_NUMBER: ClassVar[int]
    SPARK_CLUSTER_ENVIRONMENT_VERSION_FIELD_NUMBER: ClassVar[int]
    job_request_templates: JobRequestTemplates
    merged_user_deployment_settings_version: int
    spark_cluster_environment_version: int
    def __init__(self, spark_cluster_environment_version: Optional[int] = ..., job_request_templates: Optional[Union[JobRequestTemplates, Mapping]] = ..., merged_user_deployment_settings_version: Optional[int] = ...) -> None: ...

class TaskType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class SparkExecutionEnvironment(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class InstanceProfile(_message.Message):
    __slots__ = ["instance_profile_arn"]
    INSTANCE_PROFILE_ARN_FIELD_NUMBER: ClassVar[int]
    instance_profile_arn: str
    def __init__(self, instance_profile_arn: Optional[str] = ...) -> None: ...

class InstanceProfilesListResponse(_message.Message):
    __slots__ = ["instance_profiles"]
    INSTANCE_PROFILES_FIELD_NUMBER: ClassVar[int]
    instance_profiles: _containers.RepeatedCompositeFieldContainer[InstanceProfile]
    def __init__(self, instance_profiles: Optional[Iterable[Union[InstanceProfile, Mapping]]] = ...) -> None: ...



import contextlib
import logging
from typing import List
from typing import Tuple

import pyarrow
import ray

from tecton_core import offline_store
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.builder import build_materialization_querytree
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.query_tree_executor import QueryTreeExecutor
from tecton_materialization.common.task_params import TimeInterval
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.common.task_params import job_query_intervals
from tecton_materialization.ray.bulk_backfill import get_bootstrap_bulk_backfill_qt
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.materialization_utils import get_delta_writer_for_fd
from tecton_materialization.ray.materialization_utils import update_status_table
from tecton_materialization.ray.materialization_utils import write_to_online_store
from tecton_materialization.ray.nodes import AddTimePartitionNode
from tecton_proto.data.feature_view__client_pb2 import FeatureView
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.offlinestore.delta import metadata__client_pb2 as metadata_pb2


logger = logging.getLogger(__name__)

PYARROW_ASC = "ascending"
PYARROW_DESC = "descending"


def run_batch_materialization(
    materialization_task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    batch_params = materialization_task_params.batch_task_info.batch_parameters
    is_plan_integration_test_job = materialization_task_params.HasField("plan_id")
    # Bulk Backfill jobs (i.e. read_from_offline_store_for_online_write is true) don't need to break up the time
    # intervals since it reads from the offline store instead of the data source.
    if (
        batch_params.write_to_offline_feature_store
        or (batch_params.write_to_online_feature_store and not batch_params.read_from_offline_store_for_online_write)
        or is_plan_integration_test_job
    ):
        intervals = job_query_intervals(materialization_task_params)
        for idx, interval in enumerate(intervals):
            job_status_client.set_query_index(idx, len(intervals))
            if is_plan_integration_test_job:
                materialize_interval_for_plan_integration_test(
                    interval=interval,
                    materialization_task_params=materialization_task_params,
                    executor=executor,
                )
            else:
                materialize_interval(
                    interval=interval,
                    materialization_task_params=materialization_task_params,
                    job_status_client=job_status_client,
                    executor=executor,
                )

    # For Bulk Backfill, we need to export data from offline store to the intermediate store here. And then we have a
    # separate job to import the data from the intermediate store to the online store.
    should_write_to_online_from_offline_store = (
        batch_params.write_to_online_feature_store
        and batch_params.read_from_offline_store_for_online_write
        and not is_plan_integration_test_job
    )
    if should_write_to_online_from_offline_store:
        fd = feature_definition_from_task_params(materialization_task_params)
        qt = get_bootstrap_bulk_backfill_qt(
            fd, batch_params.feature_start_time.ToDatetime(), batch_params.feature_end_time.ToDatetime()
        )
        bulk_backfill_data = executor.exec_qt(qt).result_table

        bulk_backfill_stage_monitor = job_status_client.create_stage_monitor(
            TectonManagedStage.StageType.BULK_LOAD,
            "Upload Backfill data to the intermediate Bulk Load store",
        )
        with bulk_backfill_stage_monitor():
            output_arrow_data_as_json(
                bulk_backfill_data,
                materialization_task_params.batch_task_info.dynamodb_json_output_path,
                materialization_task_params.feature_view,
            )

        update_status_table(
            materialization_task_params.online_store_writer_config,
            materialization_task_params.feature_view,
            fd,
            materialization_task_params.batch_task_info.batch_parameters.feature_end_time,
        )


def output_arrow_data_as_json(bulk_backfill_data: pyarrow.RecordBatchReader, output_path: str, fv: FeatureView):
    # Encode the bulk backfill data to Json and write to the intermediate store
    while True:
        try:
            next_batch = bulk_backfill_data.read_next_batch()
        except StopIteration:
            return
        batch_bytes = serialize_pyarrow_batch(next_batch)
        output_arrow_record_batch(batch_bytes, fv, output_path)


def serialize_pyarrow_batch(next_batch) -> bytes:
    sink = pyarrow.BufferOutputStream()
    with pyarrow.ipc.new_stream(sink, next_batch.schema) as writer:
        writer.write_batch(next_batch)
    return sink.getvalue().to_pybytes()


def output_arrow_record_batch(batch_bytes: bytes, fv: FeatureView, output_path: str):
    runner_function = ray.cross_language.java_function(
        "com.tecton.onlinestorewriter.OutputArrowBatchAsDynamoJson", "runFromArrowRecordBatch"
    )
    job = runner_function.remote(batch_bytes, fv.SerializeToString(), output_path)
    ray.get(job)


def materialize_interval_for_plan_integration_test(
    interval: TimeInterval,
    materialization_task_params: MaterializationTaskParams,
    executor: QueryTreeExecutor,
):
    # skip writing to offline store to not overwrite prod data
    fd = feature_definition_from_task_params(materialization_task_params)
    qt = _get_batch_materialization_plan(fd, interval)
    materialized_data = executor.exec_qt(qt).result_table
    num_rows = sum(batch.num_rows for batch in materialized_data)
    logger.info(f"Generated {num_rows} rows for FV {fd.name} for interval from {interval.start} to {interval.end}")


def _calculate_materialized_data(
    executor: QueryTreeExecutor,
    fd: FeatureDefinitionWrapper,
    interval: TimeInterval,
) -> pyarrow.RecordBatchReader:
    qt = _get_batch_materialization_plan(fd, interval)
    materialized_data = executor.exec_qt(qt).result_table

    # Sorting rows withing batches helps improve writing parquet files: fewer partitions are written in parallel.
    # Also, secondary sorting by join keys can improve reading performance (if filter by join key will be pushed down to arrow reader).
    return sort_rows_in_batches(
        materialized_data,
        by=[(offline_store.TIME_PARTITION, PYARROW_ASC), *[(key, PYARROW_ASC) for key in fd.join_keys]],
    )


def materialize_interval(
    interval: TimeInterval,
    materialization_task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    fd = feature_definition_from_task_params(materialization_task_params)
    assert fd.writes_to_offline_store, f"Offline materialization is required for FeatureView {fd.id} ({fd.name})"
    assert fd.has_delta_offline_store, f"Delta is required for FeatureView {fd.id} ({fd.name})"

    batch_params = materialization_task_params.batch_task_info.batch_parameters
    should_write_to_online_store_from_source = (
        batch_params.write_to_online_feature_store and not batch_params.read_from_offline_store_for_online_write
    )

    is_overwrite = materialization_task_params.batch_task_info.batch_parameters.is_overwrite

    delta_writer = get_delta_writer_for_fd(materialization_task_params)
    parts = None

    transaction_metadata = metadata_pb2.TectonDeltaMetadata()
    transaction_metadata.feature_start_time.FromDatetime(interval.start)
    transaction_exists = delta_writer.transaction_exists(transaction_metadata)
    if not is_overwrite and transaction_exists:
        offline_stage_monitor = job_status_client.create_stage_monitor(
            TectonManagedStage.StageType.OFFLINE_STORE,
            f"Skipping writing to offline store. Found previous commit in range {interval.start} - {interval.end}",
        )
        with offline_stage_monitor():
            logger.info(
                f"Found previous commit with metadata {transaction_metadata} for data in range {interval.start} - {interval.end}. Skipping writing to delta table."
            )
    else:

        @delta_writer.transaction(transaction_metadata)
        def txn() -> List[str]:
            if is_overwrite:
                delta_writer.delete_time_range(interval)
            materialized_data = _calculate_materialized_data(executor, fd, interval)
            offline_stage_monitor = job_status_client.create_stage_monitor(
                TectonManagedStage.StageType.OFFLINE_STORE,
                "Unload features to offline store",
            )
            with offline_stage_monitor():
                return delta_writer.write(materialized_data)

        parts = txn()

    if should_write_to_online_store_from_source:
        online_stage_monitor = job_status_client.create_stage_monitor(
            TectonManagedStage.StageType.ONLINE_STORE,
            "Unload features to online store",
        )
        with online_stage_monitor(), contextlib.ExitStack() as stack:
            if parts is None:
                # We skipped the txn because of matching metadata, but we still need to write out the parquet files for
                # the online store writer. We can accomplish this by using write() and then later abort() to delete
                # the files when we're done with them.
                materialized_data = _calculate_materialized_data(executor, fd, interval)

                parts = delta_writer.write(materialized_data)
                stack.callback(delta_writer.abort)

            # TODO(meastham): Probably should send these all at once to the online store copier
            for uri in parts:
                write_to_online_store(
                    materialization_task_params.online_store_writer_config,
                    materialization_task_params.feature_view,
                    materialization_task_params.batch_task_info.batch_parameters.feature_end_time,
                    fd,
                    uri,
                )


def _get_batch_materialization_plan(fd: FeatureDefinitionWrapper, interval: TimeInterval) -> NodeRef:
    tree = build_materialization_querytree(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        fdw=fd,
        for_stream=False,
        feature_data_time_limits=interval.to_pendulum(),
    )
    return AddTimePartitionNode.for_feature_definition(fd, tree)


def sort_rows_in_batches(reader: pyarrow.RecordBatchReader, by: List[Tuple[str, str]]) -> pyarrow.RecordBatchReader:
    """
    Create new RecordBatchReader with rows sorted within each batch.

    :param reader: iterator over record batches
    :param by: sorting conditions - list of tuple(name, order)
    """

    def batch_iter():
        while True:
            try:
                next_batch = reader.read_next_batch()
            except StopIteration:
                return

            yield next_batch.sort_by(by)

    return pyarrow.RecordBatchReader.from_batches(reader.schema, batch_iter())



from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.spark_common import clusters__client_pb2 as _clusters__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor
JOB_METADATA_TABLE_TYPE_DYNAMO: JobMetadataTableType
JOB_METADATA_TABLE_TYPE_GCS: JobMetadataTableType
JOB_METADATA_TABLE_TYPE_UNSPECIFIED: JobMetadataTableType
OFFLINE_STORE_TYPE_DBFS: OfflineStoreType
OFFLINE_STORE_TYPE_GCS: OfflineStoreType
OFFLINE_STORE_TYPE_S3: OfflineStoreType
OFFLINE_STORE_TYPE_UNSPECIFIED: OfflineStoreType
ONLINE_STORE_TYPE_BIGTABLE: OnlineStoreType
ONLINE_STORE_TYPE_DYNAMO: OnlineStoreType
ONLINE_STORE_TYPE_REDIS: OnlineStoreType
ONLINE_STORE_TYPE_UNSPECIFIED: OnlineStoreType

class ComputeConsumptionInfo(_message.Message):
    __slots__ = ["compute_usage", "duration"]
    COMPUTE_USAGE_FIELD_NUMBER: ClassVar[int]
    DURATION_FIELD_NUMBER: ClassVar[int]
    compute_usage: _containers.RepeatedCompositeFieldContainer[ComputeUsage]
    duration: _duration_pb2.Duration
    def __init__(self, duration: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., compute_usage: Optional[Iterable[Union[ComputeUsage, Mapping]]] = ...) -> None: ...

class ComputeUsage(_message.Message):
    __slots__ = ["instance_availability", "instance_count", "instance_type"]
    INSTANCE_AVAILABILITY_FIELD_NUMBER: ClassVar[int]
    INSTANCE_COUNT_FIELD_NUMBER: ClassVar[int]
    INSTANCE_TYPE_FIELD_NUMBER: ClassVar[int]
    instance_availability: _clusters__client_pb2.AwsAvailability
    instance_count: int
    instance_type: str
    def __init__(self, instance_availability: Optional[Union[_clusters__client_pb2.AwsAvailability, str]] = ..., instance_type: Optional[str] = ..., instance_count: Optional[int] = ...) -> None: ...

class JobMetadata(_message.Message):
    __slots__ = ["materialization_consumption_info", "online_store_copier_execution_info", "spark_execution_info", "tecton_managed_info"]
    MATERIALIZATION_CONSUMPTION_INFO_FIELD_NUMBER: ClassVar[int]
    ONLINE_STORE_COPIER_EXECUTION_INFO_FIELD_NUMBER: ClassVar[int]
    SPARK_EXECUTION_INFO_FIELD_NUMBER: ClassVar[int]
    TECTON_MANAGED_INFO_FIELD_NUMBER: ClassVar[int]
    materialization_consumption_info: MaterializationConsumptionInfo
    online_store_copier_execution_info: OnlineStoreCopierExecutionInfo
    spark_execution_info: SparkJobExecutionInfo
    tecton_managed_info: TectonManagedInfo
    def __init__(self, online_store_copier_execution_info: Optional[Union[OnlineStoreCopierExecutionInfo, Mapping]] = ..., spark_execution_info: Optional[Union[SparkJobExecutionInfo, Mapping]] = ..., tecton_managed_info: Optional[Union[TectonManagedInfo, Mapping]] = ..., materialization_consumption_info: Optional[Union[MaterializationConsumptionInfo, Mapping]] = ...) -> None: ...

class MaterializationConsumptionInfo(_message.Message):
    __slots__ = ["compute_consumption", "offline_store_consumption", "online_store_consumption"]
    COMPUTE_CONSUMPTION_FIELD_NUMBER: ClassVar[int]
    OFFLINE_STORE_CONSUMPTION_FIELD_NUMBER: ClassVar[int]
    ONLINE_STORE_CONSUMPTION_FIELD_NUMBER: ClassVar[int]
    compute_consumption: ComputeConsumptionInfo
    offline_store_consumption: OfflineStoreWriteConsumptionInfo
    online_store_consumption: OnlineStoreWriteConsumptionInfo
    def __init__(self, offline_store_consumption: Optional[Union[OfflineStoreWriteConsumptionInfo, Mapping]] = ..., online_store_consumption: Optional[Union[OnlineStoreWriteConsumptionInfo, Mapping]] = ..., compute_consumption: Optional[Union[ComputeConsumptionInfo, Mapping]] = ...) -> None: ...

class OfflineConsumptionBucket(_message.Message):
    __slots__ = ["features_written", "rows_written"]
    FEATURES_WRITTEN_FIELD_NUMBER: ClassVar[int]
    ROWS_WRITTEN_FIELD_NUMBER: ClassVar[int]
    features_written: int
    rows_written: int
    def __init__(self, rows_written: Optional[int] = ..., features_written: Optional[int] = ...) -> None: ...

class OfflineStoreWriteConsumptionInfo(_message.Message):
    __slots__ = ["consumption_info", "offline_store_type"]
    class ConsumptionInfoEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: int
        value: OfflineConsumptionBucket
        def __init__(self, key: Optional[int] = ..., value: Optional[Union[OfflineConsumptionBucket, Mapping]] = ...) -> None: ...
    CONSUMPTION_INFO_FIELD_NUMBER: ClassVar[int]
    OFFLINE_STORE_TYPE_FIELD_NUMBER: ClassVar[int]
    consumption_info: _containers.MessageMap[int, OfflineConsumptionBucket]
    offline_store_type: OfflineStoreType
    def __init__(self, consumption_info: Optional[Mapping[int, OfflineConsumptionBucket]] = ..., offline_store_type: Optional[Union[OfflineStoreType, str]] = ...) -> None: ...

class OnlineConsumptionBucket(_message.Message):
    __slots__ = ["features_written", "rows_written"]
    FEATURES_WRITTEN_FIELD_NUMBER: ClassVar[int]
    ROWS_WRITTEN_FIELD_NUMBER: ClassVar[int]
    features_written: int
    rows_written: int
    def __init__(self, rows_written: Optional[int] = ..., features_written: Optional[int] = ...) -> None: ...

class OnlineStoreCopierExecutionInfo(_message.Message):
    __slots__ = ["is_revoked"]
    IS_REVOKED_FIELD_NUMBER: ClassVar[int]
    is_revoked: bool
    def __init__(self, is_revoked: bool = ...) -> None: ...

class OnlineStoreWriteConsumptionInfo(_message.Message):
    __slots__ = ["consumption_info", "online_store_type"]
    class ConsumptionInfoEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: int
        value: OnlineConsumptionBucket
        def __init__(self, key: Optional[int] = ..., value: Optional[Union[OnlineConsumptionBucket, Mapping]] = ...) -> None: ...
    CONSUMPTION_INFO_FIELD_NUMBER: ClassVar[int]
    ONLINE_STORE_TYPE_FIELD_NUMBER: ClassVar[int]
    consumption_info: _containers.MessageMap[int, OnlineConsumptionBucket]
    online_store_type: OnlineStoreType
    def __init__(self, consumption_info: Optional[Mapping[int, OnlineConsumptionBucket]] = ..., online_store_type: Optional[Union[OnlineStoreType, str]] = ...) -> None: ...

class SparkJobExecutionInfo(_message.Message):
    __slots__ = ["is_revoked", "run_id", "stream_handoff_synchronization_info"]
    IS_REVOKED_FIELD_NUMBER: ClassVar[int]
    RUN_ID_FIELD_NUMBER: ClassVar[int]
    STREAM_HANDOFF_SYNCHRONIZATION_INFO_FIELD_NUMBER: ClassVar[int]
    is_revoked: bool
    run_id: str
    stream_handoff_synchronization_info: StreamHandoffSynchronizationInfo
    def __init__(self, run_id: Optional[str] = ..., is_revoked: bool = ..., stream_handoff_synchronization_info: Optional[Union[StreamHandoffSynchronizationInfo, Mapping]] = ...) -> None: ...

class StreamHandoffSynchronizationInfo(_message.Message):
    __slots__ = ["new_cluster_started", "query_cancellation_complete", "query_cancellation_requested", "stream_query_start_allowed"]
    NEW_CLUSTER_STARTED_FIELD_NUMBER: ClassVar[int]
    QUERY_CANCELLATION_COMPLETE_FIELD_NUMBER: ClassVar[int]
    QUERY_CANCELLATION_REQUESTED_FIELD_NUMBER: ClassVar[int]
    STREAM_QUERY_START_ALLOWED_FIELD_NUMBER: ClassVar[int]
    new_cluster_started: bool
    query_cancellation_complete: bool
    query_cancellation_requested: bool
    stream_query_start_allowed: bool
    def __init__(self, new_cluster_started: bool = ..., stream_query_start_allowed: bool = ..., query_cancellation_requested: bool = ..., query_cancellation_complete: bool = ...) -> None: ...

class TectonManagedInfo(_message.Message):
    __slots__ = ["stages", "state"]
    STAGES_FIELD_NUMBER: ClassVar[int]
    STATE_FIELD_NUMBER: ClassVar[int]
    stages: _containers.RepeatedCompositeFieldContainer[TectonManagedStage]
    state: TectonManagedStage.State
    def __init__(self, stages: Optional[Iterable[Union[TectonManagedStage, Mapping]]] = ..., state: Optional[Union[TectonManagedStage.State, str]] = ...) -> None: ...

class TectonManagedStage(_message.Message):
    __slots__ = ["compiled_sql_query", "description", "duration", "error_detail", "error_type", "external_link", "progress", "stage_type", "start_time", "state"]
    class ErrorType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class StageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    AGGREGATE: TectonManagedStage.StageType
    BIGQUERY: TectonManagedStage.StageType
    BULK_LOAD: TectonManagedStage.StageType
    CANCELLED: TectonManagedStage.State
    COMPILED_SQL_QUERY_FIELD_NUMBER: ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    DURATION_FIELD_NUMBER: ClassVar[int]
    ERROR: TectonManagedStage.State
    ERROR_DETAIL_FIELD_NUMBER: ClassVar[int]
    ERROR_TYPE_FIELD_NUMBER: ClassVar[int]
    ERROR_TYPE_UNSPECIFIED: TectonManagedStage.ErrorType
    EXTERNAL_LINK_FIELD_NUMBER: ClassVar[int]
    OFFLINE_STORE: TectonManagedStage.StageType
    ONLINE_STORE: TectonManagedStage.StageType
    PENDING: TectonManagedStage.State
    PROGRESS_FIELD_NUMBER: ClassVar[int]
    PYTHON: TectonManagedStage.StageType
    RUNNING: TectonManagedStage.State
    SNOWFLAKE: TectonManagedStage.StageType
    STAGE_TYPE_FIELD_NUMBER: ClassVar[int]
    START_TIME_FIELD_NUMBER: ClassVar[int]
    STATE_FIELD_NUMBER: ClassVar[int]
    STATE_UNSPECIFIED: TectonManagedStage.State
    SUCCESS: TectonManagedStage.State
    UNEXPECTED_ERROR: TectonManagedStage.ErrorType
    USER_ERROR: TectonManagedStage.ErrorType
    compiled_sql_query: str
    description: str
    duration: _duration_pb2.Duration
    error_detail: str
    error_type: TectonManagedStage.ErrorType
    external_link: str
    progress: float
    stage_type: TectonManagedStage.StageType
    start_time: _timestamp_pb2.Timestamp
    state: TectonManagedStage.State
    def __init__(self, stage_type: Optional[Union[TectonManagedStage.StageType, str]] = ..., state: Optional[Union[TectonManagedStage.State, str]] = ..., external_link: Optional[str] = ..., progress: Optional[float] = ..., description: Optional[str] = ..., error_type: Optional[Union[TectonManagedStage.ErrorType, str]] = ..., error_detail: Optional[str] = ..., compiled_sql_query: Optional[str] = ..., duration: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., start_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...

class OnlineStoreType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class OfflineStoreType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class JobMetadataTableType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.dataobs import metric__client_pb2 as _metric__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureExpectation(_message.Message):
    __slots__ = ["alert_message_template", "creation_time", "expression", "input_column_names", "last_update_time", "name"]
    ALERT_MESSAGE_TEMPLATE_FIELD_NUMBER: ClassVar[int]
    CREATION_TIME_FIELD_NUMBER: ClassVar[int]
    EXPRESSION_FIELD_NUMBER: ClassVar[int]
    INPUT_COLUMN_NAMES_FIELD_NUMBER: ClassVar[int]
    LAST_UPDATE_TIME_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    alert_message_template: str
    creation_time: _timestamp_pb2.Timestamp
    expression: str
    input_column_names: _containers.RepeatedScalarFieldContainer[str]
    last_update_time: _timestamp_pb2.Timestamp
    name: str
    def __init__(self, name: Optional[str] = ..., expression: Optional[str] = ..., alert_message_template: Optional[str] = ..., input_column_names: Optional[Iterable[str]] = ..., creation_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., last_update_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...

class MetricExpectation(_message.Message):
    __slots__ = ["alert_message_template", "creation_time", "display_name", "expression", "input_metrics", "last_update_time", "name"]
    ALERT_MESSAGE_TEMPLATE_FIELD_NUMBER: ClassVar[int]
    CREATION_TIME_FIELD_NUMBER: ClassVar[int]
    DISPLAY_NAME_FIELD_NUMBER: ClassVar[int]
    EXPRESSION_FIELD_NUMBER: ClassVar[int]
    INPUT_METRICS_FIELD_NUMBER: ClassVar[int]
    LAST_UPDATE_TIME_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    alert_message_template: str
    creation_time: _timestamp_pb2.Timestamp
    display_name: str
    expression: str
    input_metrics: _containers.RepeatedCompositeFieldContainer[_metric__client_pb2.FeatureMetric]
    last_update_time: _timestamp_pb2.Timestamp
    name: str
    def __init__(self, name: Optional[str] = ..., display_name: Optional[str] = ..., expression: Optional[str] = ..., alert_message_template: Optional[str] = ..., input_metrics: Optional[Iterable[Union[_metric__client_pb2.FeatureMetric, Mapping]]] = ..., creation_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., last_update_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...



from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.common import analytics_options__client_pb2 as _analytics_options__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Mapping, Optional

DESCRIPTOR: _descriptor.FileDescriptor

class BasicInfo(_message.Message):
    __slots__ = ["description", "name", "owner", "tags"]
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    OWNER_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    description: str
    name: str
    owner: str
    tags: _containers.ScalarMap[str, str]
    def __init__(self, name: Optional[str] = ..., description: Optional[str] = ..., owner: Optional[str] = ..., tags: Optional[Mapping[str, str]] = ...) -> None: ...



import logging
import sys

import click
from colorama import Fore

from tecton import tecton_context
from tecton.cli import printer
from tecton.cli import repo
from tecton.cli import workspace_utils
from tecton.cli.command import TectonGroup
from tecton.cli.engine import update_tecton_state
from tecton.cli.workspace_utils import WorkspaceType
from tecton.cli.workspace_utils import switch_to_workspace


logger = logging.getLogger(__name__)

PROD_WORKSPACE_NAME = "prod"


@click.group("workspace", cls=TectonGroup, is_main_command=True)
def workspace():
    """Manage Tecton Workspaces."""


@workspace.command()
@click.argument("workspace", type=WorkspaceType())
def select(workspace):
    """Select Tecton Workspace."""
    workspace_utils.check_workspace_exists(workspace)
    switch_to_workspace(workspace)


@workspace.command()
def list():
    """List available Tecton Workspaces."""
    current_workspace = tecton_context.get_current_workspace()
    workspaces = workspace_utils.list_workspaces()
    materializable = [w.name for w in workspaces if w.capabilities.materializable]
    nonmaterializable = [w.name for w in workspaces if not w.capabilities.materializable]

    if materializable:
        printer.safe_print("Live Workspaces:")
        for name in materializable:
            marker = "*" if name == current_workspace else " "
            printer.safe_print(f"{marker} {name}")

    # Print whitespace between the two sections if needed.
    if materializable and nonmaterializable:
        printer.safe_print()

    if nonmaterializable:
        printer.safe_print("Development Workspaces:")
        for name in nonmaterializable:
            marker = "*" if name == current_workspace else " "
            printer.safe_print(f"{marker} {name}")


@workspace.command()
def show():
    """Show active Workspace."""
    workspace_name = tecton_context.get_current_workspace()
    workspace = workspace_utils.get_workspace(workspace_name)
    workspace_type = "Live" if workspace.capabilities.materializable else "Development"
    printer.safe_print(f"{workspace_name} ({workspace_type})")


@workspace.command()
@click.argument("workspace")
@click.option(
    "--live/--no-live",
    # Kept for backwards compatibility
    "--automatic-materialization-enabled/--automatic-materialization-disabled",
    default=False,
    help="Create a live Workspace, which enables materialization and online serving.",
)
def create(workspace, live):
    """Create a new Tecton Workspace"""
    # There is a check for this on the server side too, but we optimistically validate
    # here as well to show a pretty error message.
    workspace_names = {w.name for w in workspace_utils.list_workspaces()}
    if workspace in workspace_names:
        printer.safe_print(f"Workspace {workspace} already exists", file=sys.stderr)
        sys.exit(1)

    # create
    workspace_utils.create_workspace(workspace, live)

    # switch to new workspace
    switch_to_workspace(workspace)
    printer.safe_print(
        """
You have created a new, empty workspace. Workspaces let
you create and manage an isolated feature repository.
Running "tecton plan" will compare your local repository
against the remote repository, which is initially empty.
    """
    )


@workspace.command()
@click.argument("workspace", type=WorkspaceType())
@click.option("--yes", "-y", is_flag=True)
def delete(workspace, yes):
    """Delete a Tecton Workspace."""
    # validate
    if workspace == PROD_WORKSPACE_NAME:
        printer.safe_print(f"Deleting Workspace '{PROD_WORKSPACE_NAME}' not allowed.")
        sys.exit(1)

    is_live = workspace_utils.is_live_workspace(workspace)

    # confirm deletion
    confirmation = "y" if yes else None
    while confirmation not in ("y", "n", ""):
        confirmation_text = f'Are you sure you want to delete the workspace "{workspace}"? (y/N)'
        if is_live:
            confirmation_text = f"{Fore.RED}Warning{Fore.RESET}: This will delete any materialized data in this workspace.\n{confirmation_text}"
        confirmation = input(confirmation_text).lower().strip()
    if confirmation == "n" or confirmation == "":
        printer.safe_print("Cancelling delete action.")
        sys.exit(1)

    # archive all fcos in the remote state unconditionally.
    # This will need to be updated when workspaces support materialization.
    update_tecton_state(
        objects=[],
        repo_root="",
        repo_config=None,
        repo_files=[],
        apply=True,
        # Set interactive to False to avoid duplicate confirmation.
        # Confirmation of this action is handled above already.
        interactive=False,
        upgrade_all=False,
        workspace_name=workspace,
    )

    # delete
    workspace_utils.delete_workspace(workspace)

    # switch to prod if deleted current
    if workspace == tecton_context.get_current_workspace():
        switch_to_workspace(PROD_WORKSPACE_NAME)


workspace.add_command(repo.restore)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.spark_common import clusters__client_pb2 as _clusters__client_pb2
from tecton_proto.spark_common import libraries__client_pb2 as _libraries__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor
INSTANCE_ALLOCATION_FAILURE: RunTerminationReason
JOB_FINISHED: RunTerminationReason
MANUAL_CANCELATION: RunTerminationReason
NON_CLOUD_FAILURE: RunTerminationReason
RUN_STATUS_CANCELED: RunStatus
RUN_STATUS_ERROR: RunStatus
RUN_STATUS_PENDING: RunStatus
RUN_STATUS_RUNNING: RunStatus
RUN_STATUS_SUBMISSION_ERROR: RunStatus
RUN_STATUS_SUCCESS: RunStatus
RUN_STATUS_TERMINATING: RunStatus
RUN_STATUS_UNKNOWN: RunStatus
SUBMISSION_ERROR: RunTerminationReason
UNKNOWN_TERMINATION_REASON: RunTerminationReason

class GetJobRequest(_message.Message):
    __slots__ = ["run_id"]
    RUN_ID_FIELD_NUMBER: ClassVar[int]
    run_id: str
    def __init__(self, run_id: Optional[str] = ...) -> None: ...

class GetJobResponse(_message.Message):
    __slots__ = ["additional_metadata", "details", "job_id", "run_id", "run_page_url", "spark_cluster_id"]
    class AdditionalMetadataEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ADDITIONAL_METADATA_FIELD_NUMBER: ClassVar[int]
    DETAILS_FIELD_NUMBER: ClassVar[int]
    JOB_ID_FIELD_NUMBER: ClassVar[int]
    RUN_ID_FIELD_NUMBER: ClassVar[int]
    RUN_PAGE_URL_FIELD_NUMBER: ClassVar[int]
    SPARK_CLUSTER_ID_FIELD_NUMBER: ClassVar[int]
    additional_metadata: _containers.ScalarMap[str, str]
    details: RunDetails
    job_id: str
    run_id: str
    run_page_url: str
    spark_cluster_id: str
    def __init__(self, run_id: Optional[str] = ..., job_id: Optional[str] = ..., run_page_url: Optional[str] = ..., spark_cluster_id: Optional[str] = ..., details: Optional[Union[RunDetails, Mapping]] = ..., additional_metadata: Optional[Mapping[str, str]] = ...) -> None: ...

class ListJobRequest(_message.Message):
    __slots__ = ["marker", "offset"]
    MARKER_FIELD_NUMBER: ClassVar[int]
    OFFSET_FIELD_NUMBER: ClassVar[int]
    marker: str
    offset: int
    def __init__(self, offset: Optional[int] = ..., marker: Optional[str] = ...) -> None: ...

class ListJobResponse(_message.Message):
    __slots__ = ["has_more", "marker", "runs"]
    HAS_MORE_FIELD_NUMBER: ClassVar[int]
    MARKER_FIELD_NUMBER: ClassVar[int]
    RUNS_FIELD_NUMBER: ClassVar[int]
    has_more: bool
    marker: str
    runs: _containers.RepeatedCompositeFieldContainer[RunSummary]
    def __init__(self, runs: Optional[Iterable[Union[RunSummary, Mapping]]] = ..., has_more: bool = ..., marker: Optional[str] = ...) -> None: ...

class PythonMaterializationTask(_message.Message):
    __slots__ = ["base_parameters", "materialization_path_uri", "taskType"]
    class TaskType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class BaseParametersEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    BASE_PARAMETERS_FIELD_NUMBER: ClassVar[int]
    BATCH: PythonMaterializationTask.TaskType
    DATASET_GENERATION: PythonMaterializationTask.TaskType
    DELETION: PythonMaterializationTask.TaskType
    DELTA_MAINTENANCE: PythonMaterializationTask.TaskType
    FEATURE_EXPORT: PythonMaterializationTask.TaskType
    INGEST: PythonMaterializationTask.TaskType
    MATERIALIZATION_PATH_URI_FIELD_NUMBER: ClassVar[int]
    PLAN_INTEGRATION_TEST_BATCH: PythonMaterializationTask.TaskType
    PLAN_INTEGRATION_TEST_STREAM: PythonMaterializationTask.TaskType
    STREAMING: PythonMaterializationTask.TaskType
    TASKTYPE_FIELD_NUMBER: ClassVar[int]
    base_parameters: _containers.ScalarMap[str, str]
    materialization_path_uri: str
    taskType: PythonMaterializationTask.TaskType
    def __init__(self, materialization_path_uri: Optional[str] = ..., base_parameters: Optional[Mapping[str, str]] = ..., taskType: Optional[Union[PythonMaterializationTask.TaskType, str]] = ...) -> None: ...

class RunDetails(_message.Message):
    __slots__ = ["end_time", "run_status", "start_time", "state_message", "termination_reason", "vendor_termination_reason"]
    END_TIME_FIELD_NUMBER: ClassVar[int]
    RUN_STATUS_FIELD_NUMBER: ClassVar[int]
    START_TIME_FIELD_NUMBER: ClassVar[int]
    STATE_MESSAGE_FIELD_NUMBER: ClassVar[int]
    TERMINATION_REASON_FIELD_NUMBER: ClassVar[int]
    VENDOR_TERMINATION_REASON_FIELD_NUMBER: ClassVar[int]
    end_time: _timestamp_pb2.Timestamp
    run_status: RunStatus
    start_time: _timestamp_pb2.Timestamp
    state_message: str
    termination_reason: RunTerminationReason
    vendor_termination_reason: str
    def __init__(self, run_status: Optional[Union[RunStatus, str]] = ..., termination_reason: Optional[Union[RunTerminationReason, str]] = ..., state_message: Optional[str] = ..., vendor_termination_reason: Optional[str] = ..., start_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., end_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...

class RunSummary(_message.Message):
    __slots__ = ["additional_metadata", "resource_locator", "run_id", "run_state"]
    class AdditionalMetadataEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ADDITIONAL_METADATA_FIELD_NUMBER: ClassVar[int]
    RESOURCE_LOCATOR_FIELD_NUMBER: ClassVar[int]
    RUN_ID_FIELD_NUMBER: ClassVar[int]
    RUN_STATE_FIELD_NUMBER: ClassVar[int]
    additional_metadata: _containers.ScalarMap[str, str]
    resource_locator: str
    run_id: str
    run_state: str
    def __init__(self, run_id: Optional[str] = ..., run_state: Optional[str] = ..., resource_locator: Optional[str] = ..., additional_metadata: Optional[Mapping[str, str]] = ...) -> None: ...

class StartJobRequest(_message.Message):
    __slots__ = ["existing_cluster", "is_notebook", "libraries", "materialization_task", "new_cluster", "run_name", "timeout_seconds", "use_stepped_materialization"]
    EXISTING_CLUSTER_FIELD_NUMBER: ClassVar[int]
    IS_NOTEBOOK_FIELD_NUMBER: ClassVar[int]
    LIBRARIES_FIELD_NUMBER: ClassVar[int]
    MATERIALIZATION_TASK_FIELD_NUMBER: ClassVar[int]
    NEW_CLUSTER_FIELD_NUMBER: ClassVar[int]
    RUN_NAME_FIELD_NUMBER: ClassVar[int]
    TIMEOUT_SECONDS_FIELD_NUMBER: ClassVar[int]
    USE_STEPPED_MATERIALIZATION_FIELD_NUMBER: ClassVar[int]
    existing_cluster: _clusters__client_pb2.ExistingCluster
    is_notebook: bool
    libraries: _containers.RepeatedCompositeFieldContainer[_libraries__client_pb2.Library]
    materialization_task: PythonMaterializationTask
    new_cluster: _clusters__client_pb2.NewCluster
    run_name: str
    timeout_seconds: int
    use_stepped_materialization: bool
    def __init__(self, new_cluster: Optional[Union[_clusters__client_pb2.NewCluster, Mapping]] = ..., existing_cluster: Optional[Union[_clusters__client_pb2.ExistingCluster, Mapping]] = ..., materialization_task: Optional[Union[PythonMaterializationTask, Mapping]] = ..., run_name: Optional[str] = ..., libraries: Optional[Iterable[Union[_libraries__client_pb2.Library, Mapping]]] = ..., timeout_seconds: Optional[int] = ..., is_notebook: bool = ..., use_stepped_materialization: bool = ...) -> None: ...

class StartJobResponse(_message.Message):
    __slots__ = ["run_id"]
    RUN_ID_FIELD_NUMBER: ClassVar[int]
    run_id: str
    def __init__(self, run_id: Optional[str] = ...) -> None: ...

class StopJobRequest(_message.Message):
    __slots__ = ["run_id"]
    RUN_ID_FIELD_NUMBER: ClassVar[int]
    run_id: str
    def __init__(self, run_id: Optional[str] = ...) -> None: ...

class RunStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class RunTerminationReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/realtime/instance_group__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import container_image__client_pb2 as tecton__proto_dot_common_dot_container__image____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/realtime/instance_group__client.proto\x12\x15tecton_proto.realtime\x1a\x31tecton_proto/common/container_image__client.proto\"\xb0\x08\n\rInstanceGroup\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12L\n\x0f\x63ontainer_image\x18\x02 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\x0e\x63ontainerImage\x12\x41\n\x08\x63\x61pacity\x18\x03 \x01(\x0b\x32%.tecton_proto.realtime.CapacityConfigR\x08\x63\x61pacity\x12Z\n\x13health_check_config\x18\x04 \x01(\x0b\x32(.tecton_proto.realtime.HealthCheckConfigH\x00R\x11healthCheckConfig\x12,\n\x11health_check_name\x18\x0c \x01(\tH\x00R\x0fhealthCheckName\x12I\n\x0fprometheus_port\x18\x05 \x01(\x0b\x32 .tecton_proto.realtime.NamedPortR\x0eprometheusPort\x12=\n\tgrpc_port\x18\x06 \x01(\x0b\x32 .tecton_proto.realtime.NamedPortR\x08grpcPort\x12=\n\thttp_port\x18\x0b \x01(\x0b\x32 .tecton_proto.realtime.NamedPortR\x08httpPort\x12W\n\x12\x61ws_instance_group\x18\x07 \x01(\x0b\x32\'.tecton_proto.realtime.AWSInstanceGroupH\x01R\x10\x61wsInstanceGroup\x12p\n\x1bgoogle_cloud_instance_group\x18\x08 \x01(\x0b\x32/.tecton_proto.realtime.GoogleCloudInstanceGroupH\x01R\x18googleCloudInstanceGroup\x12\x42\n\x04tags\x18\t \x03(\x0b\x32..tecton_proto.realtime.InstanceGroup.TagsEntryR\x04tags\x12s\n\x15\x65nvironment_variables\x18\n \x03(\x0b\x32>.tecton_proto.realtime.InstanceGroup.EnvironmentVariablesEntryR\x14\x65nvironmentVariables\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1aG\n\x19\x45nvironmentVariablesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x13\n\x11health_check_typeB\x0c\n\ncloud_type\"\xae\x01\n\x13InstanceGroupHandle\x12*\n\x11instance_group_id\x18\x01 \x01(\tR\x0finstanceGroupId\x12.\n\x13instance_group_name\x18\x02 \x01(\tR\x11instanceGroupName\x12;\n\x1ainstance_group_template_id\x18\x03 \x01(\tR\x17instanceGroupTemplateId\"\xa0\x01\n\x0e\x43\x61pacityConfig\x12/\n\x13\x61utoscaling_enabled\x18\x04 \x01(\x08R\x12\x61utoscalingEnabled\x12\x1b\n\tmin_nodes\x18\x01 \x01(\x05R\x08minNodes\x12\x1b\n\tmax_nodes\x18\x02 \x01(\x05R\x08maxNodes\x12#\n\rdesired_nodes\x18\x03 \x01(\x05R\x0c\x64\x65siredNodes\"B\n\x13InstanceGroupStatus\x12+\n\x11healthy_instances\x18\x01 \x01(\x05R\x10healthyInstances\"\x82\x01\n\x11HealthCheckConfig\x12\x34\n\x04port\x18\x01 \x01(\x0b\x32 .tecton_proto.realtime.NamedPortR\x04port\x12\x12\n\x04path\x18\x02 \x01(\tR\x04path\x12#\n\rprotocol_type\x18\x03 \x01(\tR\x0cprotocolType\"I\n\tNamedPort\x12\x1f\n\x0bport_number\x18\x01 \x01(\x05R\nportNumber\x12\x1b\n\tport_name\x18\x02 \x01(\tR\x08portName\"\xcf\x03\n\x10\x41WSInstanceGroup\x12\x32\n\x15\x61utoscaling_group_arn\x18\x01 \x01(\tR\x13\x61utoscalingGroupArn\x12\x34\n\x16\x61utoscaling_group_name\x18\x02 \x01(\tR\x14\x61utoscalingGroupName\x12\x16\n\x06region\x18\x03 \x01(\tR\x06region\x12\x12\n\x04port\x18\x04 \x01(\x05R\x04port\x12*\n\x11health_check_path\x18\x05 \x01(\tR\x0fhealthCheckPath\x12#\n\rinstance_type\x18\x06 \x01(\tR\x0cinstanceType\x12 \n\x0c\x61mi_image_id\x18\x07 \x01(\tR\namiImageId\x12\x37\n\x18iam_instance_profile_arn\x18\x08 \x01(\tR\x15iamInstanceProfileArn\x12,\n\x12security_group_ids\x18\t \x03(\tR\x10securityGroupIds\x12\x1d\n\nsubnet_ids\x18\n \x03(\tR\tsubnetIds\x12,\n\x12launch_template_id\x18\x0b \x01(\tR\x10launchTemplateId\"e\n\x1c\x41WSInstanceGroupUpdateConfig\x12#\n\rinstance_type\x18\x01 \x01(\tR\x0cinstanceType\x12 \n\x0c\x61mi_image_id\x18\x02 \x01(\tR\namiImageId\"\x86\x01\n\x0e\x41WSTargetGroup\x12\x10\n\x03\x61rn\x18\x01 \x01(\tR\x03\x61rn\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12N\n\x0einstance_group\x18\x03 \x01(\x0b\x32\'.tecton_proto.realtime.AWSInstanceGroupR\rinstanceGroup\"\xda\x01\n\x18GoogleCloudInstanceGroup\x12\x18\n\x07project\x18\x01 \x01(\tR\x07project\x12\x16\n\x06region\x18\x02 \x01(\tR\x06region\x12\x1b\n\ttarget_id\x18\x03 \x01(\tR\x08targetId\x12!\n\x0cmachine_type\x18\x04 \x01(\tR\x0bmachineType\x12 \n\x0bsubnetworks\x18\x05 \x03(\tR\x0bsubnetworks\x12*\n\x11health_check_name\x18\x06 \x01(\tR\x0fhealthCheckName\"\xc2\x01\n\x19GoogleCloudBackendService\x12\x1b\n\ttarget_id\x18\x01 \x01(\tR\x08targetId\x12\x18\n\x07project\x18\x02 \x01(\tR\x07project\x12\x16\n\x06region\x18\x03 \x01(\tR\x06region\x12V\n\x0einstance_group\x18\x04 \x01(\x0b\x32/.tecton_proto.realtime.GoogleCloudInstanceGroupR\rinstanceGroup\"\xdb\x01\n\x12LoadBalancerTarget\x12Q\n\x10\x61ws_target_group\x18\x01 \x01(\x0b\x32%.tecton_proto.realtime.AWSTargetGroupH\x00R\x0e\x61wsTargetGroup\x12h\n\x16google_backend_service\x18\x02 \x01(\x0b\x32\x30.tecton_proto.realtime.GoogleCloudBackendServiceH\x00R\x14googleBackendServiceB\x08\n\x06targetB\x1c\n\x18\x63om.tecton.realtimeprotoP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.realtime.instance_group__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.tecton.realtimeprotoP\001'
  _INSTANCEGROUP_TAGSENTRY._options = None
  _INSTANCEGROUP_TAGSENTRY._serialized_options = b'8\001'
  _INSTANCEGROUP_ENVIRONMENTVARIABLESENTRY._options = None
  _INSTANCEGROUP_ENVIRONMENTVARIABLESENTRY._serialized_options = b'8\001'
  _INSTANCEGROUP._serialized_start=129
  _INSTANCEGROUP._serialized_end=1201
  _INSTANCEGROUP_TAGSENTRY._serialized_start=1038
  _INSTANCEGROUP_TAGSENTRY._serialized_end=1093
  _INSTANCEGROUP_ENVIRONMENTVARIABLESENTRY._serialized_start=1095
  _INSTANCEGROUP_ENVIRONMENTVARIABLESENTRY._serialized_end=1166
  _INSTANCEGROUPHANDLE._serialized_start=1204
  _INSTANCEGROUPHANDLE._serialized_end=1378
  _CAPACITYCONFIG._serialized_start=1381
  _CAPACITYCONFIG._serialized_end=1541
  _INSTANCEGROUPSTATUS._serialized_start=1543
  _INSTANCEGROUPSTATUS._serialized_end=1609
  _HEALTHCHECKCONFIG._serialized_start=1612
  _HEALTHCHECKCONFIG._serialized_end=1742
  _NAMEDPORT._serialized_start=1744
  _NAMEDPORT._serialized_end=1817
  _AWSINSTANCEGROUP._serialized_start=1820
  _AWSINSTANCEGROUP._serialized_end=2283
  _AWSINSTANCEGROUPUPDATECONFIG._serialized_start=2285
  _AWSINSTANCEGROUPUPDATECONFIG._serialized_end=2386
  _AWSTARGETGROUP._serialized_start=2389
  _AWSTARGETGROUP._serialized_end=2523
  _GOOGLECLOUDINSTANCEGROUP._serialized_start=2526
  _GOOGLECLOUDINSTANCEGROUP._serialized_end=2744
  _GOOGLECLOUDBACKENDSERVICE._serialized_start=2747
  _GOOGLECLOUDBACKENDSERVICE._serialized_end=2941
  _LOADBALANCERTARGET._serialized_start=2944
  _LOADBALANCERTARGET._serialized_end=3163
# @@protoc_insertion_point(module_scope)



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas

from tecton_core import conf
from tecton_core import specs
from tecton_core.errors import UDF_ERROR
from tecton_core.errors import UDF_TYPE_ERROR
from tecton_core.id_helper import IdHelper
from tecton_core.pipeline.feature_pipeline import FeaturePipeline
from tecton_core.pipeline.feature_pipeline import NodeValueType
from tecton_core.pipeline.pipeline_common import CONSTANT_TYPE
from tecton_core.pipeline.pipeline_common import constant_node_to_value
from tecton_core.realtime_context import RealtimeContext
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_proto.common.id__client_pb2 import Id as TectonId


class RealtimeFeaturePipeline(FeaturePipeline):
    def __init__(
        self,
        name: str,
        pipeline: Pipeline,
        transformations: List[specs.TransformationSpec],
        is_prompt: bool,
        events_df_timestamp_field: Optional[str] = None,
        pipeline_inputs: Optional[Dict[str, Union[Dict[str, Any], pandas.DataFrame, RealtimeContext]]] = None,
    ) -> None:
        self._pipeline = pipeline
        self._name = name
        self._id_to_transformation = {t.id: t for t in transformations}
        self._events_df_timestamp_field = events_df_timestamp_field
        self._is_prompt = is_prompt

        root_transformation = self.get_transformation_by_id(self._pipeline.root.transformation_node.transformation_id)
        assert root_transformation.transformation_mode in (
            TransformationMode.TRANSFORMATION_MODE_PYTHON,
            TransformationMode.TRANSFORMATION_MODE_PANDAS,
        )
        # In Spark, the UDF cannot reference a proto enum, so instead save mode as a string
        self.mode = (
            "python"
            if root_transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON
            else "pandas"
        )

        # Access this conf value outside of the UDF to avoid doing it many times and avoiding any worker/driver state issues.
        self._should_check_output_schema = conf.get_bool("TECTON_PYTHON_ODFV_OUTPUT_SCHEMA_CHECK_ENABLED")
        self._pipeline_inputs = pipeline_inputs

    def get_transformation_by_id(self, id: TectonId) -> specs.TransformationSpec:
        return self._id_to_transformation[IdHelper.to_string(id)]

    @property
    def is_pandas_mode(self):
        return self.mode == "pandas"

    @property
    def is_python_mode(self):
        return self.mode == "python"

    def _context_node_to_value(self, pipeline_node: PipelineNode) -> Optional[Union[pandas.DataFrame, RealtimeContext]]:
        assert self._pipeline_inputs is not None
        return self._pipeline_inputs[pipeline_node.context_node.input_name]

    def _request_data_node_to_value(self, pipeline_node: PipelineNode) -> Union[Dict[str, Any], pandas.DataFrame]:
        assert self._pipeline_inputs is not None
        return self._pipeline_inputs[pipeline_node.request_data_source_node.input_name]

    def _feature_view_node_to_value(self, pipeline_node: PipelineNode) -> Union[Dict[str, Any], pandas.DataFrame]:
        assert self._pipeline_inputs is not None
        return self._pipeline_inputs[pipeline_node.feature_view_node.input_name]

    def _node_to_value(self, pipeline_node: PipelineNode) -> NodeValueType:
        if pipeline_node.HasField("constant_node"):
            return self._constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("feature_view_node"):
            return self._feature_view_node_to_value(pipeline_node)
        elif pipeline_node.HasField("request_data_source_node"):
            return self._request_data_node_to_value(pipeline_node)
        elif pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_value(pipeline_node.transformation_node)
        elif pipeline_node.HasField("context_node"):
            return self._context_node_to_value(pipeline_node)
        elif pipeline_node.HasField("materialization_context_node"):
            msg = f"MaterializationContext is unsupported for {self._fco_name}s."
            raise ValueError(msg)
        else:
            msg = f"This is not yet implemented {pipeline_node}"
            raise NotImplementedError(msg)

    def _constant_node_to_value(self, pipeline_node: PipelineNode) -> CONSTANT_TYPE:
        return constant_node_to_value(pipeline_node)

    @staticmethod
    def _format_values_for_pandas_mode(input_df: pandas.DataFrame) -> pandas.DataFrame:
        for col in input_df.select_dtypes(include=["datetime64"]).columns:
            input_df[col] = pandas.to_datetime(input_df[col], utc=True)
        return input_df

    def _transformation_node_to_value(
        self, transformation_node: TransformationNode
    ) -> Union[Dict[str, Any], pandas.DataFrame]:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[Union[str, int, float, bool]] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            input_node = transformation_input.node
            node_value = self._node_to_value(input_node)

            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        return self._apply_transformation_function(transformation_node, args, kwargs)

    def _apply_transformation_function(
        self, transformation_node: TransformationNode, args: List[Any], kwargs: Dict[str, Any]
    ) -> Union[Dict[str, Any], pandas.DataFrame]:
        """For the given transformation node, returns the corresponding DataFrame transformation.

        If needed, resulted function is wrapped with a function that translates mode-specific input/output types to DataFrames.
        """
        transformation = self.get_transformation_by_id(transformation_node.transformation_id)
        mode = transformation.transformation_mode
        user_function = transformation.user_function

        if (
            mode != TransformationMode.TRANSFORMATION_MODE_PANDAS
            and mode != TransformationMode.TRANSFORMATION_MODE_PYTHON
        ):
            msg = f"Unsupported transformation mode({transformation.transformation_mode}) for {self._fco_name}s."
            raise KeyError(msg)

        try:
            resp = user_function(*args, **kwargs)
            return self._wrap_resp(resp)
        except TypeError as e:
            raise UDF_TYPE_ERROR(e)
        except Exception as e:
            raise UDF_ERROR(e, transformation.metadata.name)

    def _wrap_resp(self, resp):
        if self._is_prompt:
            if not isinstance(resp, str):
                msg = "Prompt functions must return strings"
                raise TypeError(msg)
            return {"prompt": resp}
        else:
            return resp

    @property
    def _fco_name(self):
        if self._is_prompt:
            return "Prompt"
        else:
            return "Realtime Feature View"

    def run_with_inputs(
        self, inputs: Union[Dict[str, pandas.DataFrame], Dict[str, Any]]
    ) -> Union[CONSTANT_TYPE, Dict[str, Any], pandas.DataFrame, pandas.Series]:
        self._pipeline_inputs = inputs
        return self.get_dataframe()



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Pair(_message.Message):
    __slots__ = ["first", "second"]
    FIRST_FIELD_NUMBER: ClassVar[int]
    SECOND_FIELD_NUMBER: ClassVar[int]
    first: str
    second: str
    def __init__(self, first: Optional[str] = ..., second: Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/realtime_compute__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.realtime import instance_group__client_pb2 as tecton__proto_dot_realtime_dot_instance__group____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0tecton_proto/data/realtime_compute__client.proto\x12\x11tecton_proto.data\x1a\x32tecton_proto/realtime/instance_group__client.proto\"\x18\n\x16\x43olocatedComputeConfig\"d\n\x1bRemoteFunctionComputeConfig\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12!\n\x0c\x66unction_uri\x18\x03 \x01(\tR\x0b\x66unctionUri\"\x8e\x01\n\x1aInstanceGroupComputeConfig\x12\x1d\n\ngroup_name\x18\x01 \x01(\tR\tgroupName\x12Q\n\x0einstance_group\x18\x02 \x01(\x0b\x32*.tecton_proto.realtime.InstanceGroupHandleR\rinstanceGroup\"\xb8\x02\n\x13OnlineComputeConfig\x12X\n\x11\x63olocated_compute\x18\x01 \x01(\x0b\x32).tecton_proto.data.ColocatedComputeConfigH\x00R\x10\x63olocatedCompute\x12W\n\x0eremote_compute\x18\x02 \x01(\x0b\x32..tecton_proto.data.RemoteFunctionComputeConfigH\x00R\rremoteCompute\x12\x63\n\x15instance_group_config\x18\x03 \x01(\x0b\x32-.tecton_proto.data.InstanceGroupComputeConfigH\x00R\x13instanceGroupConfigB\t\n\x07\x63omputeB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.realtime_compute__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _COLOCATEDCOMPUTECONFIG._serialized_start=123
  _COLOCATEDCOMPUTECONFIG._serialized_end=147
  _REMOTEFUNCTIONCOMPUTECONFIG._serialized_start=149
  _REMOTEFUNCTIONCOMPUTECONFIG._serialized_end=249
  _INSTANCEGROUPCOMPUTECONFIG._serialized_start=252
  _INSTANCEGROUPCOMPUTECONFIG._serialized_end=394
  _ONLINECOMPUTECONFIG._serialized_start=397
  _ONLINECOMPUTECONFIG._serialized_end=709
# @@protoc_insertion_point(module_scope)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.common import fco_locator__client_pb2 as _fco_locator__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DATA_TABLE_STATE_DELETED: OnlineDataTableState
DATA_TABLE_STATE_NOT_READY: OnlineDataTableState
DATA_TABLE_STATE_PENDING_DELETION: OnlineDataTableState
DATA_TABLE_STATE_READY: OnlineDataTableState
DATA_TABLE_STATE_UNKNOWN_UNSPECIFIED: OnlineDataTableState
DESCRIPTOR: _descriptor.FileDescriptor
ONLINE_BACKFILL_LOAD_TYPE_BULK: OnlineBackfillLoadType
ONLINE_BACKFILL_LOAD_TYPE_COMPACTION: OnlineBackfillLoadType
ONLINE_BACKFILL_LOAD_TYPE_TASK: OnlineBackfillLoadType
ONLINE_BACKFILL_LOAD_TYPE_UNSPECIFIED: OnlineBackfillLoadType
TABLE_FORMAT_VERSION_DEFAULT_UNSPECIFIED: TableFormatVersion
TABLE_FORMAT_VERSION_V2: TableFormatVersion
TABLE_FORMAT_VERSION_V3: TableFormatVersion

class FileLocation(_message.Message):
    __slots__ = ["path"]
    PATH_FIELD_NUMBER: ClassVar[int]
    path: str
    def __init__(self, path: Optional[str] = ...) -> None: ...

class FvMaterialization(_message.Message):
    __slots__ = ["feature_export_data_location", "feature_store_format_version", "has_been_revisted", "has_materialized_data", "id_feature_view_locator", "last_revisited_ts", "materialization_serial_version", "materialized_data_location", "misc_offline_data_location", "online_data_tables", "online_table_import_complete", "online_table_imported_by_materialization", "streaming_checkpoint_locations"]
    FEATURE_EXPORT_DATA_LOCATION_FIELD_NUMBER: ClassVar[int]
    FEATURE_STORE_FORMAT_VERSION_FIELD_NUMBER: ClassVar[int]
    HAS_BEEN_REVISTED_FIELD_NUMBER: ClassVar[int]
    HAS_MATERIALIZED_DATA_FIELD_NUMBER: ClassVar[int]
    ID_FEATURE_VIEW_LOCATOR_FIELD_NUMBER: ClassVar[int]
    LAST_REVISITED_TS_FIELD_NUMBER: ClassVar[int]
    MATERIALIZATION_SERIAL_VERSION_FIELD_NUMBER: ClassVar[int]
    MATERIALIZED_DATA_LOCATION_FIELD_NUMBER: ClassVar[int]
    MISC_OFFLINE_DATA_LOCATION_FIELD_NUMBER: ClassVar[int]
    ONLINE_DATA_TABLES_FIELD_NUMBER: ClassVar[int]
    ONLINE_TABLE_IMPORTED_BY_MATERIALIZATION_FIELD_NUMBER: ClassVar[int]
    ONLINE_TABLE_IMPORT_COMPLETE_FIELD_NUMBER: ClassVar[int]
    STREAMING_CHECKPOINT_LOCATIONS_FIELD_NUMBER: ClassVar[int]
    feature_export_data_location: FileLocation
    feature_store_format_version: int
    has_been_revisted: bool
    has_materialized_data: bool
    id_feature_view_locator: _fco_locator__client_pb2.IdFcoLocator
    last_revisited_ts: _timestamp_pb2.Timestamp
    materialization_serial_version: int
    materialized_data_location: FileLocation
    misc_offline_data_location: FileLocation
    online_data_tables: _containers.RepeatedCompositeFieldContainer[OnlineDataTable]
    online_table_import_complete: bool
    online_table_imported_by_materialization: bool
    streaming_checkpoint_locations: _containers.RepeatedCompositeFieldContainer[FileLocation]
    def __init__(self, id_feature_view_locator: Optional[Union[_fco_locator__client_pb2.IdFcoLocator, Mapping]] = ..., materialization_serial_version: Optional[int] = ..., materialized_data_location: Optional[Union[FileLocation, Mapping]] = ..., streaming_checkpoint_locations: Optional[Iterable[Union[FileLocation, Mapping]]] = ..., has_materialized_data: bool = ..., feature_store_format_version: Optional[int] = ..., misc_offline_data_location: Optional[Union[FileLocation, Mapping]] = ..., online_table_imported_by_materialization: bool = ..., online_table_import_complete: bool = ..., feature_export_data_location: Optional[Union[FileLocation, Mapping]] = ..., online_data_tables: Optional[Iterable[Union[OnlineDataTable, Mapping]]] = ..., has_been_revisted: bool = ..., last_revisited_ts: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...

class OnlineDataTable(_message.Message):
    __slots__ = ["batch_table_format_version", "feature_data_watermark", "name", "state_transitions"]
    BATCH_TABLE_FORMAT_VERSION_FIELD_NUMBER: ClassVar[int]
    FEATURE_DATA_WATERMARK_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    STATE_TRANSITIONS_FIELD_NUMBER: ClassVar[int]
    batch_table_format_version: TableFormatVersion
    feature_data_watermark: _timestamp_pb2.Timestamp
    name: str
    state_transitions: _containers.RepeatedCompositeFieldContainer[OnlineDataTableStateTransition]
    def __init__(self, name: Optional[str] = ..., feature_data_watermark: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., state_transitions: Optional[Iterable[Union[OnlineDataTableStateTransition, Mapping]]] = ..., batch_table_format_version: Optional[Union[TableFormatVersion, str]] = ...) -> None: ...

class OnlineDataTableStateTransition(_message.Message):
    __slots__ = ["state", "timestamp"]
    STATE_FIELD_NUMBER: ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: ClassVar[int]
    state: OnlineDataTableState
    timestamp: _timestamp_pb2.Timestamp
    def __init__(self, state: Optional[Union[OnlineDataTableState, str]] = ..., timestamp: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...

class OnlineBackfillLoadType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class TableFormatVersion(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class OnlineDataTableState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from datetime import timedelta
from pathlib import Path
from unittest import TestCase
from unittest import mock

from tecton import RedshiftConfig
from tecton import v09_compat
from tecton.cli.upgrade_utils import BatchFeatureViewGuidance
from tecton.cli.upgrade_utils import DataSourceGuidance
from tecton.cli.upgrade_utils import _get_imports_for_type
from tecton.types import Array
from tecton.types import Field
from tecton.types import Int64
from tecton.types import Map
from tecton.types import String
from tecton.types import Struct
from tecton.types import Timestamp
from tecton_core import conf


class TestUpgrade(TestCase):
    def mockPatch(self, *args, **kwargs):
        patcher = mock.patch(*args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def setUp(self) -> None:
        conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "spark")
        self.mock_metadata_service = mock.MagicMock()
        self.mockPatch("tecton._internals.metadata_service.instance", return_value=self.mock_metadata_service)

    def test_push_source(self):
        push_source = v09_compat.PushSource(
            name="PushSource",
            schema=[
                Field(name="content_keyword", dtype=String),
                Field(name="timestamp", dtype=Timestamp),
                Field(name="clicked", dtype=Int64),
            ],
        )

        push_source_guidance = DataSourceGuidance(push_source, str(Path()))._get_upgrade_guidance()
        self.assertEqual(len(push_source_guidance), 2)
        self.assertIn(
            "Replace `PushSource` with `StreamSource` and set `stream_config=PushConfig()", push_source_guidance[0]
        )
        self.assertIn("Update import from `tecton.v09_compat` to `tecton`.", push_source_guidance[1])

    def test_data_source(self):
        batch_source = v09_compat.BatchSource(
            name="redshift_ds", batch_config=RedshiftConfig(endpoint="test_uri", table="test_table")
        )
        push_source_guidance = DataSourceGuidance(batch_source, str(Path()))._get_upgrade_guidance()
        self.assertEqual(len(push_source_guidance), 1)
        self.assertIn(
            "No code change needed - just update import from `tecton.v09_compat` to `tecton`.", push_source_guidance[0]
        )

    def test_filtered_source(self):
        batch_source = v09_compat.BatchSource(
            name="redshift_ds", batch_config=RedshiftConfig(endpoint="test_uri", table="test_table")
        )
        entity = v09_compat.Entity(name="user", join_keys=["user_id"])

        @v09_compat.batch_feature_view(
            sources=[v09_compat.FilteredSource(batch_source)],
            entities=[entity],
            mode="spark_sql",
            batch_schedule=timedelta(hours=1),
            run_transformation_validation=False,
            timestamp_field="timestamp",
            schema=[Field("user_id", String), Field("event_id", String), Field("timestamp", Timestamp)],
        )
        def v09_compat_bfv(filtered_source):
            pass

        with mock.patch("tecton.cli.upgrade_utils._get_feature_view", return_value=v09_compat_bfv):
            bfv_guidance = BatchFeatureViewGuidance(v09_compat_bfv, str(Path()))._get_upgrade_guidance()
            self.assertEqual(len(bfv_guidance), 5)
            filtered_source_guidance = bfv_guidance[2]
            self.assertIn(
                "Replace data sources:\n```\nsources=[FilteredSource(filtered_source)]\n```\nwith: \n```\nsources=[filtered_source]\n```",
                filtered_source_guidance,
            )

    def test_aggregation(self):
        batch_source = v09_compat.BatchSource(
            name="redshift_ds", batch_config=RedshiftConfig(endpoint="test_uri", table="test_table")
        )
        entity = v09_compat.Entity(name="user", join_keys=["user_id"])

        @v09_compat.batch_feature_view(
            sources=[batch_source],
            entities=[entity],
            mode="spark_sql",
            batch_schedule=timedelta(hours=1),
            run_transformation_validation=False,
            timestamp_field="timestamp",
            aggregations=[
                v09_compat.Aggregation(
                    name="aggy_ag", column="event_id", function="count", time_window=timedelta(days=1)
                ),
                v09_compat.Aggregation(column="event_id", function="count", time_window=timedelta(days=2)),
            ],
            schema=[Field("user_id", String), Field("event_id", String), Field("timestamp", Timestamp)],
        )
        def v09_compat_bfv_aggregation(batch_source):
            pass

        with mock.patch("tecton.cli.upgrade_utils._get_feature_view", return_value=v09_compat_bfv_aggregation):
            bfv_guidance = BatchFeatureViewGuidance(v09_compat_bfv_aggregation, str(Path()))._get_upgrade_guidance()
            self.assertEqual(len(bfv_guidance), 6)
            self.assertIn(
                'Aggregate(name="aggy_ag", input_column=Field("event_id", String), function="count", time_window=timedelta(days=1))',
                bfv_guidance[2],
            )
            self.assertIn(
                'Aggregate(input_column=Field("event_id", String), function="count", time_window=timedelta(days=2))',
                bfv_guidance[2],
            )

    def test_get_imports_for_type(self):
        test_cases = [
            (Array(String), {"Array", "String"}),
            (Array(Array(String)), {"Array", "String"}),
            (Struct([Field("string", String)]), {"String", "Field", "Struct"}),
            (Struct([Field("string", Array(String))]), {"String", "Array", "Field", "Struct"}),
            (Map(String, Array(Int64)), {"Array", "String", "Map", "Int64"}),
        ]

        for test_case in test_cases:
            imports = _get_imports_for_type(test_case[0])
            assert imports == test_case[1]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/server_group_status__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5tecton_proto/common/server_group_status__client.proto\x12\x13tecton_proto.common*\xfd\x01\n\x11ServerGroupStatus\x12#\n\x1fSERVER_GROUP_STATUS_UNSPECIFIED\x10\x00\x12 \n\x1cSERVER_GROUP_STATUS_CREATING\x10\x01\x12\x1f\n\x1bSERVER_GROUP_STATUS_PENDING\x10\x06\x12\x1d\n\x19SERVER_GROUP_STATUS_READY\x10\x02\x12 \n\x1cSERVER_GROUP_STATUS_UPDATING\x10\x03\x12\x1d\n\x19SERVER_GROUP_STATUS_ERROR\x10\x04\x12 \n\x1cSERVER_GROUP_STATUS_DELETING\x10\x05\x42\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.server_group_status__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _SERVERGROUPSTATUS._serialized_start=79
  _SERVERGROUPSTATUS._serialized_end=332
# @@protoc_insertion_point(module_scope)



from tecton_proto.databricks_api import jobs__client_pb2 as _jobs__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor
FAILED: LibraryStatus
INSTALLED: LibraryStatus
INSTALLING: LibraryStatus
RESOLVING: LibraryStatus
RESTORED: LibraryStatus
SKIPPED: LibraryStatus
UNINSTALL_ON_RESTART: LibraryStatus

class ClusterStatusResponse(_message.Message):
    __slots__ = ["library_statuses"]
    LIBRARY_STATUSES_FIELD_NUMBER: ClassVar[int]
    library_statuses: _containers.RepeatedCompositeFieldContainer[LibraryStatusEntry]
    def __init__(self, library_statuses: Optional[Iterable[Union[LibraryStatusEntry, Mapping]]] = ...) -> None: ...

class LibraryInstallRequest(_message.Message):
    __slots__ = ["cluster_id", "libraries"]
    CLUSTER_ID_FIELD_NUMBER: ClassVar[int]
    LIBRARIES_FIELD_NUMBER: ClassVar[int]
    cluster_id: str
    libraries: _containers.RepeatedCompositeFieldContainer[_jobs__client_pb2.RemoteLibrary]
    def __init__(self, cluster_id: Optional[str] = ..., libraries: Optional[Iterable[Union[_jobs__client_pb2.RemoteLibrary, Mapping]]] = ...) -> None: ...

class LibraryStatusEntry(_message.Message):
    __slots__ = ["library", "status"]
    LIBRARY_FIELD_NUMBER: ClassVar[int]
    STATUS_FIELD_NUMBER: ClassVar[int]
    library: _jobs__client_pb2.RemoteLibrary
    status: LibraryStatus
    def __init__(self, library: Optional[Union[_jobs__client_pb2.RemoteLibrary, Mapping]] = ..., status: Optional[Union[LibraryStatus, str]] = ...) -> None: ...

class LibraryStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import importlib
from pathlib import Path
from typing import List

from tecton.framework import model_config
from tecton.framework.model_config import ModelConfig


def get_custom_models(model_config_file_path: Path) -> List[ModelConfig]:
    spec = importlib.util.spec_from_file_location(model_config_file_path.stem, model_config_file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return list(model_config._CUSTOM_MODELS)



from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class Acl(_message.Message):
    __slots__ = ["api_key_ids", "okta_ids", "permission"]
    API_KEY_IDS_FIELD_NUMBER: ClassVar[int]
    OKTA_IDS_FIELD_NUMBER: ClassVar[int]
    PERMISSION_FIELD_NUMBER: ClassVar[int]
    api_key_ids: _containers.RepeatedCompositeFieldContainer[_id__client_pb2.Id]
    okta_ids: _containers.RepeatedScalarFieldContainer[str]
    permission: str
    def __init__(self, permission: Optional[str] = ..., api_key_ids: Optional[Iterable[Union[_id__client_pb2.Id, Mapping]]] = ..., okta_ids: Optional[Iterable[str]] = ...) -> None: ...
